/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.core;

import java.util.Map;
import java.util.Set;

/**
 * Interface for factories used to created middleware-specific sensors or
 * actuators.
 * 
 * Every implementation of this interface must provide a default constructor.
 * 
 * @author jwienke
 */
public interface CoreObjectFactory {

    /**
     * Class encapsulating a sensor to configure.
     * 
     * @author jwienke
     */
    public class SensorToConfigure {

        private String key;
        private Class<?> dataTypeClass;
        private Class<? extends Sensor<?>> sensorClass;
        private Map<String, String> sensorOptions;

        /**
         * Unique key as specified by the user of the sensor.
         * 
         * @return unique key
         */
        public String getKey() {
            return key;
        }

        /**
         * Sets the unique key specified by the user for this sensor.
         * 
         * @param key
         *            new key
         */
        public void setKey(String key) {
            this.key = key;
        }

        /**
         * Returns the produced data type of the configured sensor.
         * 
         * @return data type class
         */
        public Class<?> getDataTypeClass() {
            return dataTypeClass;
        }

        /**
         * Sets the data type produced by the configured sensor.
         * 
         * @param dataTypeClass
         *            data type class that will be produced
         */
        public void setDataTypeClass(Class<?> dataTypeClass) {
            this.dataTypeClass = dataTypeClass;
        }

        /**
         * Returns the implementing class of the sensor.
         * 
         * @return implementing class of the sensor
         */
        public Class<? extends Sensor<?>> getSensorClass() {
            return sensorClass;
        }

        /**
         * Sets the implementing class of the sensor.
         * 
         * @param sensorClass
         *            implenting class
         */
        public void setSensorClass(Class<? extends Sensor<?>> sensorClass) {
            this.sensorClass = sensorClass;
        }

        /**
         * Returns a key-value style map of user specified options for this
         * sensor configuration.
         * 
         * @return options
         */
        public Map<String, String> getSensorOptions() {
            return sensorOptions;
        }

        /**
         * Sets the user specified map of options for this sensor configuration.
         * 
         * @param sensorOptions
         *            options
         */
        public void setSensorOptions(Map<String, String> sensorOptions) {
            this.sensorOptions = sensorOptions;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "SensorToConfigure [dataTypeClass=" + dataTypeClass
                    + ", key=" + key + ", sensorClass=" + sensorClass
                    + ", sensorOptions=" + sensorOptions + "]";
        }

    }

    /**
     * Class encapsulating an actuator to configure.
     * 
     * @author jwienke
     */
    public class ActuatorToConfigure {

        private String key;
        private Class<? extends Actuator> actuatorClass;
        private Class<? extends Actuator> interfaceClass;
        private Map<String, String> actuatorOptions;

        /**
         * Unique key as specified by the user of the sensor.
         * 
         * @return unique key
         */
        public String getKey() {
            return key;
        }

        /**
         * Sets the unique key specified by the user for this sensor.
         * 
         * @param key
         *            new key
         */
        public void setKey(String key) {
            this.key = key;
        }

        /**
         * Returns the implementing class of the actuator.
         * 
         * @return implementing class
         */
        public Class<? extends Actuator> getActuatorClass() {
            return actuatorClass;
        }

        /**
         * Sets the implementing class of the actuator.
         * 
         * @param actuatorClass
         *            implementing class
         */
        public void setActuatorClass(Class<? extends Actuator> actuatorClass) {
            this.actuatorClass = actuatorClass;
        }

        /**
         * Returns the middleware-independent interface implemented by the
         * actuator.
         * 
         * @return implemented interface
         */
        public Class<? extends Actuator> getInterfaceClass() {
            return interfaceClass;
        }

        /**
         * Sets the middleware-independent interface implemented by the
         * actuator.
         * 
         * @param interfaceClass
         *            implemented interface
         */
        public void setInterfaceClass(Class<? extends Actuator> interfaceClass) {
            this.interfaceClass = interfaceClass;
        }

        /**
         * Returns a key-value style map of user specified options for this
         * actuator configuration.
         * 
         * @return options
         */
        public Map<String, String> getActuatorOptions() {
            return actuatorOptions;
        }

        /**
         * Sets the user specified options for this actuator configuration.
         * 
         * @param actuatorOptions
         *            options
         */
        public void setActuatorOptions(Map<String, String> actuatorOptions) {
            this.actuatorOptions = actuatorOptions;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return "ActuatorToConfigure [actuatorClass=" + actuatorClass
                    + ", actuatorOptions=" + actuatorOptions
                    + ", interfaceClass=" + interfaceClass + ", key=" + key
                    + "]";
        }

    }

    /**
     * Exception indicating an error while creating a new instance of a managed
     * object.
     * 
     * @author jwienke
     */
    public class CoreObjectCreationException extends RuntimeException {

        /**
         * Serial version UID.
         */
        private static final long serialVersionUID = 1884532122904715170L;

        /**
         * Constructor.
         * 
         * @param message
         *            exception description.
         */
        public CoreObjectCreationException(String message) {
            super(message);
        }

        /**
         * Constructor.
         * 
         * @param message
         *            exception description
         * @param cause
         *            exception cause
         */
        public CoreObjectCreationException(String message, Throwable cause) {
            super(message, cause);
        }

        /**
         * Constructor.
         * 
         * @param cause
         *            exception cause
         */
        public CoreObjectCreationException(Throwable cause) {
            super(cause);
        }

    }

    /**
     * Indicating an error initializing a {@link CoreObjectFactory}.
     * 
     * @author jwienke
     */
    public class InitializationException extends RuntimeException {

        /**
         * Serial version UID.
         */
        private static final long serialVersionUID = 1324236151542361611L;

        /**
         * Constructor.
         */
        public InitializationException() {
            super();
        }

        /**
         * Constructor.
         * 
         * @param message
         *            exception description
         * @param cause
         *            exception cause
         */
        public InitializationException(String message, Throwable cause) {
            super(message, cause);
        }

        /**
         * Constructor.
         * 
         * @param message
         *            exception description
         */
        public InitializationException(String message) {
            super(message);
        }

        /**
         * Constructor.
         * 
         * @param cause
         *            exception cause
         */
        public InitializationException(Throwable cause) {
            super(cause);
        }

    }

    /**
     * Configures the factory with user-supplied options. This is the first
     * method invoked
     * 
     * @param options
     *            options provided by the user as key-value-pairs.
     * @throws IllegalArgumentException
     *             this exception may be thrown to indicate an error in the
     *             options provided by the user. If this is thrown, the factory
     *             may be in an undetermined state and should not be used for
     *             creating Bonsai objects
     * @throws InitializationException
     *             a general error initializing this factory occurred. It cannot
     *             be used after this exception
     */
    void initialize(Map<String, String> options)
            throws IllegalArgumentException, InitializationException;

    /**
     * Configure all sensors requested by the user and, if possible, already
     * connect them to reduce concrete sensor creation time.
     * 
     * @param sensors
     *            list of sensors
     * @throws IllegalArgumentException
     *             error in the configuration provided by the use
     * @throws CoreObjectCreationException
     *             error connecting one of the requested sensors
     */
    void configureSensors(Set<SensorToConfigure> sensors)
            throws IllegalArgumentException, CoreObjectCreationException;

    /**
     * Configure all actuators requested by the user and, if possible, already
     * connect them to reduce concrete sensor creation time.
     * 
     * @param actuators
     *            list of actuators
     * @throws IllegalArgumentException
     *             error in the configuration provided by the use
     * @throws CoreObjectCreationException
     *             error connecting one of the requested actuators
     */
    void configureActuators(Set<ActuatorToConfigure> actuators)
            throws IllegalArgumentException, CoreObjectCreationException;

    /**
     * Factory method to create a new sensor for the given data type provided by
     * the resulting sensor.
     * 
     * @param <T>
     *            data type of the sensor to create
     * @param key
     *            key under which the sensor was registered at a call to
     *            {@link #configureSensors(Set)}
     * @param dataType
     *            class object representing the data type of the sensor to
     *            create
     * @return new instance of the sensor
     * @throws IllegalArgumentException
     *             there is no sensor for the given data type with the provided
     *             key
     * @throws CoreObjectCreationException
     *             error while creating the requested sensor
     */
    <T> Sensor<T> createSensor(String key, Class<T> dataType)
            throws IllegalArgumentException, CoreObjectCreationException;

    /**
     * Tells if this factory can create a sensor of the given key with the data
     * type to return.
     * 
     * @param key
     *            key under which the sensor was registered at a call to
     *            {@link #configureSensors(Set)}
     * @param dataType
     *            class object representing the data type of the sensor to
     *            create
     * @return <code>true</code> if a sensor with this key and the specified
     *         return type can be created in a call to
     *         {@link #createSensor(String, Class)}, else <code>false</code>
     */
    boolean canCreateSensor(String key, Class<?> dataType);

    /**
     * Factory method to create a new actuator of the provided type.
     * 
     * @param <T>
     *            type of the actuator to create (minimally implemented
     *            interface)
     * @param key
     *            key under which the actuator was registered at a call to
     *            {@link #configureActuators(Set)}
     * @param actuatorClass
     *            class object representing the type of the actuator to create
     * @return new instance of the actuator
     * @throws IllegalArgumentException
     *             there is no actuator for the given type with the provided key
     * @throws CoreObjectCreationException
     *             error while creating the requested actuator
     */
    <T extends Actuator> T createActuator(String key, Class<T> actuatorClass)
            throws IllegalArgumentException, CoreObjectCreationException;

    /**
     * Tells if this factory can create an actuator with the provided key that
     * is assignable to an instance of the given interface class.
     * 
     * @param key
     *            key of the actuator
     * @param actuatorClass
     *            interface that the actuator to create must minimally fulfill
     * @return <code>true</code> if such an actuator can be created, else
     *         <code>false</code>
     */
    boolean canCreateActuator(String key,
            Class<? extends Actuator> actuatorClass);

}
