/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.selftest;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.core.BonsaiManager;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.util.helper.GoalReachedWaiter;
import de.unibi.airobots.btl.data.navigation.GlobalPlannerStateChange;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData;
import de.unibi.airobots.btl.data.navigation.PositionData;
import de.unibi.airobots.btl.data.navigation.SpeedData;
import de.unibi.airobots.btl.units.AngleUnit;
import de.unibi.airobots.btl.units.LengthUnit;

import java.io.IOException;
import java.util.ArrayList;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;

/**
 * A self test for the navigation actuator. This aims at a short driving
 * procedure of the robot to test the function of the NavigationActuator.
 * 
 * @author fsiepman
 */
public class NavigationSelfTest {

	private static final Logger LOGGER = Logger
			.getLogger(NavigationSelfTest.class);
	private static final long GOALTIME = 20000; // msec

	private NavigationActuator navAct;
	private Sensor<PositionData> posSensor;
	private Sensor<SpeedData> speedSensor;
	private Sensor<GlobalPlannerStateChange> gplanSensor;
	private SpeechActuator speechActuator;
	private ArrayList<NavigationGoalData> goalList;

	public NavigationSelfTest() {

		BasicConfigurator.configure();
		LOGGER.info("Creating NavigationSelfTest");

		// configuring BonSAI
		try {
			BonsaiManager
					.getInstance()
					.configure(
							"/vol/robocup/trunk/etc/bonsai_config/navigationselftest.xml");
		} catch (Exception e) {
			LOGGER.error("CONFIG ERROR", e);
		}

		LOGGER.info("Creating NavigationActuator");
		navAct = BonsaiManager.getInstance().createActuator(
				"NavigationActuator", NavigationActuator.class);
		LOGGER.info("Creating speechActuator");
		speechActuator = BonsaiManager.getInstance().createActuator(
				"SpeechActuator", SpeechActuator.class);
		LOGGER.info("Creating positionSensor");
		posSensor = BonsaiManager.getInstance().createSensor("PositionSensor",
				PositionData.class);
		LOGGER.info("Creating speedSensor");
		speedSensor = BonsaiManager.getInstance().createSensor("SpeedSensor",
				SpeedData.class);
		LOGGER.info("Creating gplanSensor");
		gplanSensor = BonsaiManager.getInstance().createSensor(
				"GlobalPlannerStateChangeSensor",
				GlobalPlannerStateChange.class);

		LOGGER.info("creating goalList");
		goalList = new ArrayList<NavigationGoalData>();
		goalList.add(new NavigationGoalData("NavigationSelfTest", new Double(
				"0"), new Double("1"), 0, NavigationGoalData.GoalType.GLOBAL,
				LengthUnit.METER, AngleUnit.RADIAN));
		goalList.add(new NavigationGoalData("NavigationSelfTest", new Double(
				"0"), new Double("2"), 0, NavigationGoalData.GoalType.GLOBAL,
				LengthUnit.METER, AngleUnit.RADIAN));
		goalList.add(new NavigationGoalData("NavigationSelfTest", new Double(
				"-1"), new Double("2"), 0, NavigationGoalData.GoalType.GLOBAL,
				LengthUnit.METER, AngleUnit.RADIAN));
		goalList.add(new NavigationGoalData("NavigationSelfTest", new Double(
				"0"), new Double("0"), 0, NavigationGoalData.GoalType.GLOBAL,
				LengthUnit.METER, AngleUnit.RADIAN));
		LOGGER.info("...done.");

	}

	public void startTest() {
		try {
			LOGGER.info("calling manualStop()");
			navAct.manualStop();
			LOGGER.info("creating GoalReachedWaiter");
			GoalReachedWaiter waiter = new GoalReachedWaiter(posSensor,
					speedSensor, gplanSensor, navAct);
			LOGGER.info("calling first goal");
			navAct.setGoal(goalList.get(0));
			LOGGER.info("waiting to reach first goal");
			GoalReachedWaiter.ReachStatus status = waiter.waitForReachingGoal(
					goalList.get(0), GOALTIME);

			if (status.equals(GoalReachedWaiter.ReachStatus.REACHED)) {
				navAct.manualStop();
				LOGGER.info("goal 1 REACHED");
				speechActuator.sayAccentuated("goal one was reached");
			} else {
				navAct.manualStop();
				LOGGER.info("goal could not be reached, trying the next.");
			}

			LOGGER.info("calling second goal");
			navAct.setGoal(goalList.get(1));
			LOGGER.info("waiting to reach second goal");
			status = waiter.waitForReachingGoal(goalList.get(1), GOALTIME);

			if (status.equals(GoalReachedWaiter.ReachStatus.REACHED)) {
				navAct.manualStop();
				LOGGER.info("goal 2 REACHED");
				speechActuator.sayAccentuated("goal two was reached");
			} else {
				navAct.manualStop();
				LOGGER.info("goal could not be reached, trying the next.");
			}

			LOGGER.info("calling third goal");
			navAct.setGoal(goalList.get(2));
			LOGGER.info("waiting to reach third goal");
			status = waiter.waitForReachingGoal(goalList.get(2), GOALTIME);

			if (status.equals(GoalReachedWaiter.ReachStatus.REACHED)) {
				navAct.manualStop();
				LOGGER.info("goal 3 REACHED");
				speechActuator.sayAccentuated("goal three was reached");
			} else {
				navAct.manualStop();
				LOGGER.info("goal could not be reached, trying the next.");
			}

			LOGGER.info("calling last goal");
			navAct.setGoal(goalList.get(3));
			LOGGER.info("waiting to reach fourth goal");
			status = waiter.waitForReachingGoal(goalList.get(3), GOALTIME);

			if (status.equals(GoalReachedWaiter.ReachStatus.REACHED)) {
				navAct.manualStop();
				LOGGER.info("goal 4 REACHED");
				speechActuator.sayAccentuated("goal four was reached");
			} else {
				navAct.manualStop();
				LOGGER.info("goal could not be reached. I quit.");
			}

		} catch (IOException ex) {
			LOGGER.error("could not stop");
			LOGGER.debug(ex.getStackTrace().toString());
			System.exit(-1);
		}

	}

	public static void main(String[] args) {

		LOGGER.info("Starting NavigationSelfTest");
		try {
			NavigationSelfTest test = new NavigationSelfTest();
			System.out.println("successfully created test");
			LOGGER.info("starting test now");
			test.startTest();

		} catch (Exception ex) {
			LOGGER.error("Error doing NavigationSelfTest.");
			LOGGER.debug(ex.getStackTrace().toString());
			System.exit(-1);
		}

		LOGGER.info("test complete.");
		System.exit(0);
	}
}
