/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills;

import de.unibi.airobots.bonsai.core.Actuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.skills.abstractstates.AbstractState;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.*;
import org.apache.commons.scxml.SCXMLExecutor;
import org.apache.commons.scxml.SCXMLListener;
import org.apache.commons.scxml.TriggerEvent;
import org.apache.commons.scxml.env.SimpleDispatcher;
import org.apache.commons.scxml.env.SimpleErrorReporter;
import org.apache.commons.scxml.env.jexl.JexlEvaluator;
import org.apache.commons.scxml.model.*;
import org.apache.log4j.Logger;

/**
 * <p>
 * This class demonstrates one approach for providing the base functionality
 * needed by classes representing stateful entities, whose behaviors are defined
 * via SCXML documents.
 * </p>
 * 
 * <p>
 * SCXML documents (more generically, UML state chart diagrams) can be used to
 * define stateful behavior of objects, and Commons SCXML enables developers to
 * use this model directly into the corresponding code artifacts. The resulting
 * artifacts tend to be much simpler, embody a useful separation of concerns and
 * are easier to understand and maintain. As the size of the modeled entity
 * grows, these benefits become more apparent.
 * </p>
 * 
 * <p>
 * This approach functions by registering an SCXMLListener that gets notified
 * onentry, and calls the namesake method for each state that has been entered.
 * </p>
 * 
 * <p>
 * This class swallows all exceptions only to log them. Developers of subclasses
 * should think of themselves as &quot;component developers&quot; catering to
 * other end users, and therefore ensure that the subclasses are free of
 * <code>ModelException</code>s and the like. Most methods are
 * <code>protected</code> for ease of subclassing.
 * </p>
 * 
 */
public class SkillStateMachine implements SCXMLListener {

    /**
     * The state machine that will drive the instances of this class.
     */
    private SCXML scxml = null;

    /**
     * The instance specific SCXML engine.
     */
    private SCXMLExecutor scxmlExecutor = null;

    /**
     * The instance specific SensorActuatorManager.
     */
    private SensorActuatorManager saManager = null;

    /**
     * The log.
     */
    private Logger logger = Logger.getLogger(this.getClass());
    
    /**
     * The decoder to decode the SCXML file.
     */
    private SCXMLDecoder scxmlDekoder = new SCXMLDecoder();
    
    /**
     * Payload of last transition.
     */
    private Object payload = null;
    
    /**
     * Hash map with active states.
     */
    private HashMap<String, AbstractState> activeStates = 
            new HashMap<String, AbstractState>();

    
    /**
     * Convenience constructor.
     * 
     * @param sensorActuatorManager The SensorActuatorManager to work on.
     */
    public SkillStateMachine(SensorActuatorManager sensorActuatorManager) {
        saManager = sensorActuatorManager;
    }
    
    
    /**
     * Initialize the state machine. Decodes SCXML task file.
     * 
     * @param pathToTask Path to SCXML task file.
     * @return <code>true</code> on success, <code>false</code> otherwise.
     */
    public boolean initalize(String pathToTask) {
        boolean decodingSuccessfull = 
                scxmlDekoder.parseAndValidateSCXML(pathToTask);
        
        if (!decodingSuccessfull) {
            logger.error("Error while decoding SCXML file.");
            return decodingSuccessfull;
        }
        
        scxml = scxmlDekoder.getScxml();
        setupEngine(scxml);
        
        return decodingSuccessfull;
    }
    
    
    /**
     * Instantiate and initialize the underlying executor instance.
     * 
     * @param stateMachine
     *            The state machine
     */
    private void setupEngine(final SCXML stateMachine) {
        scxmlExecutor = new SCXMLExecutor(new JexlEvaluator(), 
                new SimpleDispatcher(),
                new SimpleErrorReporter());
        scxmlExecutor.setStateMachine(stateMachine);
        scxmlExecutor.addListener(stateMachine, this);
    }

  
    /**
     * Fire an event on the SCXML engine.
     * 
     * @param event The event name.
     * @param payload Any object that will be attached to the signal.
     * @return Whether the state machine has reached a &quot;final&quot;
     *         configuration.
     */
    public boolean fireEvent(final String event, Object payload) {
        TriggerEvent evts =  new TriggerEvent(
                event, TriggerEvent.SIGNAL_EVENT, payload);
        try {
            scxmlExecutor.triggerEvent(evts);
        } catch (ModelException me) {
            logError(me);
        }
        return scxmlExecutor.getCurrentStatus().isFinal();
    }

    
    /**
     * Invoke the corresponding class of a state (same name).
     * 
     * @param className The name of the class (equal state name).
     * @return True. if invoke was successful.
     */
    public boolean invoke(final String className) {
        logger.info("Invoking state/class: " + className);
        try {
            Class<?> c =
                    Class.forName("de.unibi.airobots.bonsai.skills.states."
                            + className);

            Constructor<?>[] constrs = c.getConstructors();
            Constructor<?> aConstructor = constrs[0];

            AbstractState aState;
            aState = (AbstractState) aConstructor.newInstance(this);
            
            activeStates.put(className, aState);
            activeStates.get(className).start();

            return true;
        } catch (InstantiationException e) {
            logger.error("Error while starting a State!: " + e.getMessage());
        } catch (ClassNotFoundException e) {
            logError(e);
        } catch (IllegalArgumentException e) {
            logError(e);
        } catch (IllegalAccessException e) {
            logError(e);
        } catch (InvocationTargetException e) {
            logError(e);
        }
        return false;
    }

    
    /**
     * Starts the state machine.
     */
    public void startMachine() {
        try {
            scxmlExecutor.go();
        } catch (ModelException me) {
            logError(me);
        }
    }

    
    /**
     * Reset the state machine.
     * 
     * @return Whether the reset was successful.
     */
    public boolean resetMachine() {
        try {
            scxmlExecutor.reset();
        } catch (ModelException me) {
            logError(me);
            return false;
        }
        return true;
    }

    
    /**
     * Utility method for logging error.
     * 
     * @param exception
     *            The exception leading to this error condition.
     */
    protected void logError(final Exception exception) {
        logger.error(exception.getMessage(), exception);
    }


    /**
     * Invoke the corresponding class of a state.
     * 
     * {@inheritDoc}
     */
    @Override
    public void onEntry(final TransitionTarget entered) {
        // Invoke state
        if (!(entered instanceof Parallel) && entered.getId() != null) {
            invoke(entered.getId());
        }
    }

    
    /**
     * {@inheritDoc}
     */
    @Override
    public void onTransition(final TransitionTarget from,
            final TransitionTarget to, final Transition transition) {        
        payload = scxmlExecutor.getRootContext().get("_eventdata");
        if (payload != null) {
            logger.debug("Payload detected: " + 
                    payload.getClass().getSimpleName());
        }
    }

    
    /**
     * {@inheritDoc}
     */
    @Override
    public void onExit(final TransitionTarget exited) {
        String exitedID;
        // Remove state from active states hash map
        if (!(exited instanceof Parallel) && 
                (exitedID = exited.getId()) != null) {
            if (activeStates.containsKey(exitedID)) {
                AbstractState stateToEnd = activeStates.get(exitedID);
                if (!stateToEnd.hasBeenExecuted()) {
                    logger.info("Waiting for state/class to end: " + exitedID);
                    stateToEnd.forceEnd();
                }
                activeStates.remove(exitedID);
            }
            
        }
    }
    

    /**
     * Returns a list of currently possible events.
     * 
     * @return List of Strings with names of the possible events
     */
    public LinkedList<String> getEventList() {
        LinkedList<String> list = new LinkedList<String>();
        if (scxmlExecutor.getCurrentStatus().getStates().iterator().hasNext()) {
            List<Transition> a =
                    ((State) scxmlExecutor.getCurrentStatus().getStates().iterator()
                            .next()).getTransitionsList();
            for (Iterator<Transition> iterator = a.iterator(); iterator.hasNext();) {
                String aString = iterator.next().getEvent().toString();
                list.push(aString);
            }
        }
        return list;
    }

    
    /**
     * Returns the instance of a sensor which is managed in the
     * SCXMLSensorActuatorManager.
     * 
     * @param <T> Type of the requested sensor.
     * @param sensorName Name of the sensor.
     * @return The requested sensor or null, if no sensor with the given
     * name exists.
     */
    public <T extends Sensor<?>> T getManagedSensor(String sensorName) {
        T sensorToReturn = saManager.<T>getManagedSensor(sensorName);
        if (sensorToReturn == null) {
            logger.warn("Request for a non-existent sensor: " + sensorName);
        }
        return sensorToReturn;
    }

    
    /**
     * Returns the instance of an actuator which is managed in the
     * SCXMLSensorActuatorManager.
     * 
     * @param <T> Type of the requested actuator.
     * @param actuatorName Name of the actuator.
     * @return The requested actuator or null, if no actuator with the given
     * name exists.
     */
    public <T extends Actuator> T getManagedActuator(String actuatorName) {
        T actuatorToReturn = saManager.<T>getManagedActuator(actuatorName);
        if (actuatorToReturn == null) {
            logger.warn("Request for a non-existent actuator: " + actuatorName);
        }
        return actuatorToReturn;
    }
    
    
    /**
     * List of all initialized sensors.
     * @return String list of all initialized sensors.
     */
    public String[] getSensors() {
        return saManager.getSensors();
    }
    
    
    /**
     * List of all initialized actuators.
     * @return String list of all initialized actuators.
     */
    public String[] getActuators() { 
        return saManager.getActuators(); 
    }
    
    
    /**
     * Returns the payload of the last transition.
     * @return Payload of the last transition or null, 
     * if no payload was attached.
     */
    public Object getPayload() {
        return payload;
    }
    
    
    /**
     * Add a listener to the document root.
     * 
     * @param listener SCXMLListener to be added. 
     */
    public void addListener(SCXMLListener listener) {
        if (scxmlExecutor != null) {
            scxmlExecutor.addListener(scxml, listener);
        }
    }
}
