/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.abstractstates;


import de.unibi.airobots.bonsai.core.Actuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.EndException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.ExecuteException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.InitException;
import org.apache.log4j.Logger;

/**
 * Abstract definition of a state.
 * 
 * @author nkoester
 * @author lkettenb
 */
public abstract class AbstractState extends Thread {
    /**
     * Defines the possible exit statuses of a state.
     */
    public enum ExitStatus {
        /**
         * The state finished successfully. 
         * 
         * Opportunity to define a processing status.
         * @see AbstractState#setProcessingStatus(java.lang.String) 
         */
        SUCCESS     ("success"), 
        /**
         * The state finished with a known error e.g., 
         * missing or not initialized data.
         * 
         * Opportunity to define a processing status.
         * @see AbstractState#setProcessingStatus(java.lang.String) 
         */
        ERROR       ("error"), 
        /**
         * The state finished with a unknown error e.g., IOException, 
         * NullPointerException, etc. 
         * 
         * No opportunity to define a processing status.
         */
        FAIL        ("fail");
        
        /**
         * Value/name of the this element.
         */
        private final String value;
        
        /**
         * Constructor. Sets the value.
         * 
         * @param value Value/name of the this element.
         */
        private ExitStatus(String value) {
            this.value = value;
        }
        
        /**
         * Get the value of a element. Usually the name e.g., 
         * SUCCESS: value="success".
         * 
         * @return Value of this element.
         */
        public String getValue() {
            return value;
        }
    }
    
    /**
     * The SkillStateMachine.
     */
    private SkillStateMachine stateMachine;
    
    /**
     * The Logger.
     */
    private Logger logger = Logger.getLogger(this.getClass());
    
    /**
     * Use this variable to indicate a more meaningful exits/processing status.
     * 
     * By initializing this variable the exit status will look like the 
     * following:
     * 
     * <code>successful.ContentOfThisVariable</code>
     * 
     * <p>
     * The processing status <code>wait</code> is reserved. Use it to recall the 
     * method <code>execute()</code> (e.g. if waiting for sensor data).
     * </p>
     * 
     * <p>
     * Be careful! You have to adapt all SCXML files that are using this state,
     * if you change this variable afterwards.
     * </p>
     */
    private String processingStatus = null;
    
    
    /**
     * Use this variable to attach any object (e.g. PersonManager) to the 
     * "successful" event. This payload can be used by the following state.
     * 
     * @deprecated Use should try to write any payload into the memory.
     * @see BtlMemoryActuator
     * 
     */
    @Deprecated
    private Object payload = null;
    
    /**
     * Use this status to recall the execute method.
     */
    private static final String WAIT_STATUS = "wait";
    
    /**
     * Indicates if this state was forced to end.
     */
    private boolean forceEnd = false;
    
    /**
     * Indicates if this state has been executed.
     */
    private boolean hasBeenExecuted = false;
    
    /**
     * Current exit status.
     */
    private ExitStatus exitStatus;

    /**
     * Constructor.
     * @param aStateMachine The state machine to use.
     */
    public AbstractState(SkillStateMachine aStateMachine) {
        stateMachine = aStateMachine;
    }

    
    /**
     * {@inheritDoc}
     */
    @Override
    public void run() {
        String simpleName = this.getClass().getSimpleName();
        
        // Invoke init()
        logger.info(simpleName + " invoke init().");
        exitStatus = invokeInit();
        logger.info(simpleName + " init(): " + exitStatus.getValue());
        
        // Invoke execute(), if init was successful
        if (exitStatus == ExitStatus.SUCCESS) {
            logger.info(simpleName + " invoke execute().");
            exitStatus = invokeExecute();
            logger.info(simpleName + " execute(): " + exitStatus.getValue());
        }
        
        // Invoke end() (always!)
        logger.info(simpleName + " invoke end.");
        switch (exitStatus) {
            case ERROR:
                ExitStatus endExitStatus = invokeEnd();
                if (endExitStatus == ExitStatus.FAIL) {
                    exitStatus = endExitStatus;
                } 
                break;
            case FAIL: 
                invokeEnd(); 
                break;
            default: 
                exitStatus = invokeEnd();
        }     
        logger.info(simpleName + " end(): " + exitStatus.getValue());
        
        // Mark as executed
        setExecuted();
        
        // Fire event
        if (exitStatus == ExitStatus.FAIL) {
            stateMachine.fireEvent(
                    simpleName + "." + exitStatus.getValue(), null);
        } else {
            stateMachine.fireEvent(
                    simpleName + "." + getProcessingStatus(), payload);
        }
    }
    
    
    /**
     * This method invokes the <code>init()</code> method of this state.
     * 
     * @return Exit status returned by <code>init()</code> method.
     * 
     * @see AbstractState#init() 
     * @see ExitStatus
     */
    private ExitStatus invokeInit() {
        try {
            return init();
        } catch (InitException e) {
            logger.error(e.getMessage());
            return ExitStatus.FAIL;
        } 
    }
    
    
    /**
     * This method invokes the <code>execute()</code> method of this state.
     * 
     * @return Exit status returned by <code>execute()</code> method.
     * 
     * @see AbstractState#execute() 
     * @see ExitStatus
     */
    private ExitStatus invokeExecute() {
        try {
            // Invoke execute() as long as processing status is equals "wait"
            // and the state did not fail.
            ExitStatus tmpExitStatus;
            do {
                setProcessingStatus(null);
                tmpExitStatus = execute();
            } while (WAIT_STATUS.equals(processingStatus) 
                    && tmpExitStatus == ExitStatus.SUCCESS
                    && !isForcedToEnd());
            return tmpExitStatus;
        } catch (ExecuteException e) {
            logger.error(e.getMessage());
            return ExitStatus.FAIL;
        }
    }
    
    
    /**
     * This method invokes the <code>end()</code> method of this state.
     * 
     * @return Exit status returned by <code>end()</code> method.
     * 
     * @see AbstractState#end() 
     * @see ExitStatus
     */
    private ExitStatus invokeEnd() {
        try {  
            return end();
        } catch (EndException e) {
            logger.error(e.getMessage());
            return ExitStatus.FAIL;
        }
    }
    
    
    /**
     * Sets up the initial references, parameters, actuators and sensors 
     * required by a state.
     * 
     * @return Exit status as defined in {@link ExitStatus}.
     * @throws InitException forces the skill to (<code>fail</code>) 
     * ({@link ExitStatus}).
     * 
     * @see de.unibi.airobots.bonsai.skills.states.exceptions.InitException
     * @see ExitStatus
     */
    public abstract ExitStatus init() throws InitException;

    
    /**
     * 
     * Main function of a state that contains the actual logic and (sub-) skill.
     * 
     * Please note that the processing status will be set to <code>null</code> 
     * before this method is invoked.
     * 
     * @return Exit status as defined in {@link ExitStatus}.
     * @throws ExecuteException forces the skill to (<code>fail</code>) 
     * ({@link ExitStatus}).
     * 
     * @see de.unibi.airobots.bonsai.skills.states.exceptions.ExecuteException
     */
    public abstract ExitStatus execute() throws ExecuteException;

    
    /**
     * Function that is called after the main functional part of a state has
     * been executed. Use this function to clean up.
     * 
     * This function is called even when the state failed. 
     * 
     * The information (state failed or not) is saved in protected the 
     * variable {@see #stateFailed}.
     * 
     * 
     * @return Exit status as defined in {@link ExitStatus}.
     * @throws EndException forces the skill to (<code>fail</code>) 
     * ({@link ExitStatus}).
     * 
     * @see de.unibi.airobots.bonsai.skills.states.exceptions.EndException
     */
    public abstract ExitStatus end() throws EndException;
    
    
    /**
     * Use this method to use the main logger.
     * @return The main logger initialized by the abstract state.
     */
    protected Logger getLogger() {
        return logger;
    }
    
    
    /**
     * Use this method to set a more meaningful exits status (if necessary).
     * 
     * @see AbstractState#processingStatus
     * 
     * @param processingStatus Any string.
     */
    protected void setProcessingStatus(String processingStatus) {
        this.processingStatus = processingStatus;
    }
    
    
    /**
     * Check if variable <code>exitStatus</code> was initialized.
     * 
     * @see AbstractState#processingStatus
     * 
     * @return The exit status.
     */
    private String getProcessingStatus() {
        if (processingStatus != null) {
            return (exitStatus.getValue() + "." + processingStatus);
        }
        return exitStatus.getValue();
    }    
    
    /**
     * Set payload for next transition/event.
     * @param payload Payload for following transition/event (any object).
     * 
     * @deprecated Use should try to write any payload into the memory.
     * @see BtlMemoryActuator
     */
    @Deprecated
    protected void setPayload(Object payload) {
        this.payload = payload;
    }
    
    
    /**
     * Returns payload from former state. Do not confuse this payload with
     * the payload you can attach to the next transition/event.
     * 
     * @see AbstractState#setPayload(java.lang.Object) 
     * @see SkillStateMachine#getPayload() 
     * 
     * @return Payload from former state
     *
     * @deprecated Use should try to write any payload into the memory.
     * @see BtlMemoryActuator
     */
    @Deprecated
    protected Object getPayload() {
        return stateMachine.getPayload();
    }
    
    
    /**
     * Force state to end?
     * 
     * @see AbstractState#forceEnd()
     * 
     * @return True if <code>forceEnd()</code> was called, false otherwise.
     */
    private synchronized boolean isForcedToEnd() {
        return forceEnd;
    }
    
    
    /**
     * Use this exit status to recall the execute method.
     * 
     * @return Exit status to recall the execute method.
     */
    protected String getWaitProcessingMessage() {
        return WAIT_STATUS;
    }
    
    
    /**
     * Force this state to end as soon as possible.
     */
    public synchronized void forceEnd() {
        forceEnd = true;
        try {
            wait();
        } catch (InterruptedException ex) {
            logger.error("Error while waiting for thread: " + ex.toString());
        }
    }
    
    
    /**
     * Indicates if this state has been executed.
     * 
     * @return True if this state has been executed, false otherwise.
     */
    public synchronized boolean hasBeenExecuted() {
        return hasBeenExecuted;
    }
    
    
    /**
     * Marks the state as executed and calls all waiting threads.
     */
    private synchronized void setExecuted() {
        hasBeenExecuted = true;
        notifyAll();
    }
    
    /**
     * Returns the current exit status.
     * 
     * @see ExitStatus
     * 
     * @return The current exit status.
     */
    protected ExitStatus getCurrentExitStatus() {
        return exitStatus;
    }
    
    
    /**
     * Returns the instance of an actuator which is managed in the
     * SCXMLSensorActuatorManager.
     * 
     * @param <T> Type of the requested actuator.
     * @param actuatorName Name of the actuator.
     * @return The requested actuator or null, if no actuator with the given
     * name exists.
     */
    public <T extends Actuator> T getManagedActuator(String actuatorName) {
        return stateMachine.<T>getManagedActuator(actuatorName);
    }
    
    
    /**
     * Returns the instance of a sensor which is managed in the
     * SCXMLSensorActuatorManager.
     * 
     * @param <T> Type of the requested sensor.
     * @param sensorName Name of the sensor.
     * @return The requested sensor or null, if no sensor with the given
     * name exists.
     */
    public <T extends Sensor<?>> T getManagedSensor(String sensorName) {
        return stateMachine.<T>getManagedSensor(sensorName);
    }
}