/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.helper;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.sensors.FaceIdentificationSensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.bonsai.strategies.NavigationStrategy;
import de.unibi.airobots.bonsai.util.helper.GoalReachedWaiter;
import de.unibi.airobots.btl.data.navigation.GlobalPlannerStateChange;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData;
import de.unibi.airobots.btl.data.navigation.PositionData;
import de.unibi.airobots.btl.data.navigation.SpeedData;
import de.unibi.airobots.btl.data.person.FaceIdentificationList;
import de.unibi.airobots.btl.data.person.PersonData;
import de.unibi.airobots.btl.units.AngleUnit;
import de.unibi.airobots.btl.units.LengthUnit;

import java.io.IOException;
import java.util.*;
import org.apache.log4j.Logger;

/**
 * A class responsible for the various management activities related to persons
 * found by the robot.
 * 
 * @author jwienke
 * @author lkettenb
 */
public final class PersonManager {
	/**
	 * The log.
	 */
	private static final Logger LOGGER = Logger.getLogger(PersonManager.class);

	/**
	 * Max. distance to accept a person standing in front of the robot.
	 */
	private static final double FRONT_MAX_DISTANCE = 1500;

	/**
	 * Maximum angle from the front orientation of the robot in which a person
	 * is accepted as standing in front of ToBI. This angle is meant as an
	 * absolute value to each direction.
	 */
	private static final double FRONT_MAX_ANGLE = (Math.PI / 3.0);

	/**
	 * Maximum time to recognize person.
	 */
	private static final long RECOGNITION_TIME = 20000;

	/**
	 * The yaw tolerance factor to calculate yaw tolerance.
	 * 
	 * @see de.unibi.airobots.btl.data.NavigationGoalData
	 *      #setYawTolerance(double)
	 */
	private static final int YAW_TOLERANCE_FACTOR = 5;

	/**
	 * Tries to recognize a previously stored person and returns its
	 * ('caiwicat') ID.
	 * 
	 * @param faceSensor
	 *            FaceIdentificationSensor to use.
	 * @return The ('caiwicat') ID of the recognized person. -1 of no person was
	 *         recognized.
	 * @throws IOException
	 * 
	 * @see FaceIdentificationSensor
	 */
	public static int recognize(FaceIdentificationSensor faceSensor)
			throws IOException {
		faceSensor.activate();

		// resetting the majority vote for the recognition
		faceSensor.reset();

		long startTime = System.currentTimeMillis();
		long maxTime = startTime + RECOGNITION_TIME;
		Map<Integer, Integer> recognitionCounts = new HashMap<Integer, Integer>();

		// collect recognition results
		while (System.currentTimeMillis() < maxTime) {
			long maxWaitTime = maxTime - System.currentTimeMillis();
			if (maxWaitTime < 0) {
				break;
			}
			try {
				FaceIdentificationList results = faceSensor
						.readLast(maxWaitTime);

				if (results == null) {
					continue;
				}

				// iterate over all faces classified
				for (FaceIdentificationList.FaceIdentification ident : results) {
					if (ident.isIdentified()) {
						// count how often each id was founds
						if (!recognitionCounts.containsKey(ident.getClassId())) {
							recognitionCounts.put(ident.getClassId(), 0);
						}
						recognitionCounts.put(ident.getClassId(),
								recognitionCounts.get(ident.getClassId()) + 1);
					}
				}
			} catch (InterruptedException e) {
				LOGGER.warn("Ignoring interruption", e);
				continue;
			}
		}

		faceSensor.deactivate();

		// evaluate results
		if (recognitionCounts.isEmpty()) {
			// no one recognized
			return (-1);
		}

		int maxCount = -1;
		int maxClassId = -1;
		for (Map.Entry<Integer, Integer> classCount : recognitionCounts
				.entrySet()) {
			if (classCount.getValue() > maxCount) {
				maxCount = classCount.getValue();
				maxClassId = classCount.getKey();
			}
		}

		LOGGER.info("Recognized person ID: " + maxClassId);
		return maxClassId;

	}

	/**
	 * Tells if the given person is standing in front of the robot so that it
	 * can be recognized using the camera.
	 * 
	 * @param p
	 *            Person to check
	 * @return <code>true</code> if person is standing in front of the robot
	 *         facing it within a small distance, else <code>false</code>
	 */
	public static boolean isStandingInFront(PersonData p) {
		boolean isSmallId = p.getId() >= 0;

		if (!isSmallId) {
			LOGGER.debug("illegal id: " + p.getId());
			return false;
		}

		if (p.getDistance(LengthUnit.MILLIMETER) > FRONT_MAX_DISTANCE) {
			LOGGER.debug("Person too far away");
			return false;
		}

		if (Math.cos(p.getAngle(AngleUnit.RADIAN)) < Math.cos(FRONT_MAX_ANGLE)) {
			LOGGER.debug("Person not in right angle");
			return false;
		}

		LOGGER.debug("Person is in front");

		return true;
	}

	/**
	 * This method tries to drive to a given person.
	 * 
	 * @param personData
	 *            Person data with position of the person.
	 * @param navigationActuator
	 *            The navigation actuator to use.
	 * @param positionSensor
	 *            The position sensor to use.
	 * @param speedSensor
	 *            The speed sensor to use.
	 * @param gplanChangeSensor
	 *            The global planner state change sensor to use.
	 * 
	 * @return <code>true</code> if goal was reached, <code>false</code>
	 *         otherwise
	 */
	public static boolean driveToPerson(PersonData personData,
			NavigationActuator navigationActuator,
			Sensor<PositionData> positionSensor, Sensor<SpeedData> speedSensor,
			Sensor<GlobalPlannerStateChange> gplanChangeSensor) {
		try {
			PositionData robotPositionData = positionSensor.readLast();
		} catch (IOException ex) {
			LOGGER.error("IOException while trying to read position data: "
					+ ex.toString());
		} catch (InterruptedException ex) {
			LOGGER.error("Interrupted Exception while trying to read "
					+ "position data: " + ex.toString());
		}
		PositionData personPositionData = personData.getPositionData();

		NavigationStrategy navigationStrategy = new NavigationStrategy(
				navigationActuator, positionSensor, speedSensor,
				gplanChangeSensor);
		NavigationGoalData target = new NavigationGoalData(personPositionData);
		target.setYawTolerance(Math.PI / YAW_TOLERANCE_FACTOR, AngleUnit.RADIAN);

		try {
			GoalReachedWaiter.ReachStatus reachStatus = navigationStrategy
					.driveTo(target);
			if (reachStatus.compareTo(GoalReachedWaiter.ReachStatus.REACHED) == 0) {
				return true;
			} else {
				LOGGER.error("Error while trying to reach goal.");
				return false;
			}
		} catch (IOException e) {
			LOGGER.error("Error while trying to reach goal: " + e.getMessage());
			return false;
		}
	}

	/**
	 * Returns the nearest person in front of the robot that is facing the robot
	 * ready to be recognized or anything else.
	 * 
	 * @param personSensor
	 *            Person sensor to use.
	 * 
	 * @return Person in front of robot or <code>null</code> if no person is
	 *         standing in front.
	 */
	public static PersonData getNextPersonInFront(
			Sensor<PersonList> personSensor) {

		try {
			PersonList persons = personSensor.readLast();

			if (persons.size() == 0) {
				return null;
			}

			java.util.List<PersonData> personList = new LinkedList<PersonData>();
			for (PersonData p : persons) {
				boolean isSmallId = p.getId() >= 0;
				if (isSmallId && isStandingInFront(p)) {
					personList.add(p);
				}
			}

			sortPersonsByDistance(personList);

			if (personList.isEmpty()) {
				return null;
			}

			return personList.get(0);

		} catch (InterruptedException ex) {
			LOGGER.error(ex.getMessage());
		} catch (IOException ex) {
			LOGGER.error(ex.getMessage());
		}
		return null;
	}

	/**
	 * Sorts a list of persons by distance in place, so that the person with the
	 * smallest distance to the robot is the first element of the list.
	 * 
	 * @param persons
	 *            List of persons to sort.
	 */
	private static void sortPersonsByDistance(java.util.List<PersonData> persons) {
		Collections.sort(persons, new Comparator<PersonData>() {

			@Override
			public int compare(PersonData o1, PersonData o2) {
				return new Double(o1.getDistance(LengthUnit.MILLIMETER))
						.compareTo(o2.getDistance(LengthUnit.MILLIMETER));
			}

		});
	}
}
