/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.view;

import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import javax.swing.*;
import org.apache.log4j.Logger;

/**
 * A simple panel with file chooser for a task and a BonSAI configuration file.
 * 
 * @author lkettenb
 */
public class ConfigAndTaskLoadPanel extends JPanel implements ActionListener {
    /**
     * Action command fired when 'load' button is pressed (task file).
     */
    public static final String AC_LOAD_TASK_FILE = "AC_LOAD_TASK_FILE";
    
    /**
     * Action command fired when 'load' button is pressed 
     * (BonSAI configuration file).
     */
    public static final String AC_LOAD_CONFIG_FILE = "AC_LOAD_CONFIG_FILE";
    
    /*
     * Constant for coordinates, sizes, etc.
     */
    private static final int TEXT_FIELD_SIZE = 50;
    
    /*
     * Default paths to xml files.
     */
    private String defaultConfigFilePath = "/";
    private String defaultTaskFilePath = defaultConfigFilePath;
    
    /*
     * Text fields used by this panel.
     */
    private JTextField tfBonsaiConfigFile = 
            new JTextField(defaultConfigFilePath, TEXT_FIELD_SIZE);
    private JTextField tfTaskFile = 
            new JTextField(defaultTaskFilePath, TEXT_FIELD_SIZE);
    
    /*
     * Lables used by this panel.
     */
    private static final JLabel L_BONSAI_CONFIG_FILE = 
            new JLabel("BonSAI configuration: ");
    private static final JLabel L_TASK_FILE = 
            new JLabel("Task SCXML file: ");
    
    /*
     * Buttons used by this panel.
     */
    private JButton bBrowseConfigFile = new JButton("Browse ...");
    private JButton bLoadConfigFile = new JButton("Load");
    private JButton bEditConfigFile = new JButton("Edit");
    private JButton bBrowseTaskFile = new JButton(bBrowseConfigFile.getText());
    private JButton bLoadTaskFile = new JButton(bLoadConfigFile.getText());
    private JButton bEditTaskFile = new JButton(bEditConfigFile.getText());
    
    /**
     * This action listener will be notified if 
     * the task/configuration file is ready to load.
     */
    private ActionListener actionListener;
    
    /**
     * The log.
     */
    private static final Logger LOGGER = 
            Logger.getLogger(ConfigAndTaskLoadPanel.class);
    
    /**
     * 
     * Convenience constructor.
     * 
     * @param actionListener ActionListener used by this panel.
     * @param pathToConfig Default path to configuration file.
     * @param pathToTask Default path to task file
     * 
     * @see ConfigAndTaskLoadPanel#AC_LOAD_CONFIG_FILE
     * @see ConfigAndTaskLoadPanel#AC_LOAD_TASK_FILE
     * @see ConfigAndTaskLoadPanel#actionListener
     */
    public ConfigAndTaskLoadPanel(ActionListener actionListener, 
            String pathToConfig, String pathToTask) {
        super();
        this.actionListener = actionListener;
        tfBonsaiConfigFile.setText(pathToConfig);
        tfTaskFile.setText(pathToTask);
        defaultConfigFilePath = pathToConfig;
        defaultTaskFilePath = pathToTask;
        init();
    }
    
    /**
     * Initializes everything.
     */
    private void init() {
        setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
        
        JPanel pConfig = new JPanel(new FlowLayout(FlowLayout.RIGHT));
        JPanel pTask = new JPanel(new FlowLayout(FlowLayout.RIGHT));
        
        bBrowseConfigFile.addActionListener(this);
        bBrowseTaskFile.addActionListener(this);
        bLoadConfigFile.addActionListener(actionListener);
        bLoadConfigFile.setActionCommand(AC_LOAD_CONFIG_FILE);
        bLoadTaskFile.addActionListener(actionListener);
        bLoadTaskFile.setActionCommand(AC_LOAD_TASK_FILE);
        bEditConfigFile.addActionListener(this);
        bEditTaskFile.addActionListener(this);
        
        L_TASK_FILE.setEnabled(false);
        tfTaskFile.setEnabled(false);
        bBrowseTaskFile.setEnabled(false);
        bLoadTaskFile.setEnabled(false);
        
        pConfig.add(L_BONSAI_CONFIG_FILE);
        pConfig.add(tfBonsaiConfigFile);
        pConfig.add(bBrowseConfigFile);
        pConfig.add(bLoadConfigFile);
        pConfig.add(bEditConfigFile);
        
        pTask.add(L_TASK_FILE);
        pTask.add(tfTaskFile);
        pTask.add(bBrowseTaskFile);
        pTask.add(bLoadTaskFile);
        pTask.add(bEditTaskFile);
        
        add(pConfig);
        add(pTask);
        
        setBorder(BorderFactory.createTitledBorder(" Configuration "));
    }

    @Override
    public void actionPerformed(ActionEvent e) {
        if (e.getSource() == bBrowseConfigFile) {
            final JFileChooser fc = new JFileChooser(defaultConfigFilePath);
            fc.setFileFilter(new XmlFileFilter());
            fc.setAcceptAllFileFilterUsed(false);
            if (fc.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
                tfBonsaiConfigFile.setText(fc.getSelectedFile().toString());
            }
        }
        if (e.getSource() == bBrowseTaskFile) {
            final JFileChooser fc = new JFileChooser(defaultTaskFilePath);
            fc.setFileFilter(new XmlFileFilter());
            fc.setAcceptAllFileFilterUsed(false);
            if (fc.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
                tfTaskFile.setText(fc.getSelectedFile().toString());
            }
        }
        if (e.getSource() == bEditConfigFile) {
            openEditor(getPathToConfigFile());
        }
        if (e.getSource() == bEditTaskFile) {
            openEditor(getPathToTaskFile());
        }
    }
    
    /**
     * Opens text editor 'gedit'.
     * 
     * @param path Path to document.
     */
    private void openEditor(String path) {
        try {
            Runtime.getRuntime().exec("gedit " + path);
        } catch (IOException ex) {
            LOGGER.warn("Unabled to open editor: " + ex.toString());
        }
    }
    
    
    /**
     * Get path to current task file.
     * 
     * @return path to current task file.
     */
    public String getPathToTaskFile() {
        return tfTaskFile.getText();
    }
    
    /**
     * Get path to current BonSAI configuration file.
     * 
     * @return Path to current BonSAI configuration file.
     */
    public String getPathToConfigFile() {
        return tfBonsaiConfigFile.getText();
    }
    
    /**
     * 
     * Enable/disable the task file chooser (disabled by default).
     *
     * @param enable <code>true</code> to enable, <code>false</code> otherwise.
     */
    public void setTaskFileChooserEnabled(boolean enable) {
        L_TASK_FILE.setEnabled(enable);
        tfTaskFile.setEnabled(enable);
        bBrowseTaskFile.setEnabled(enable);
        bLoadTaskFile.setEnabled(enable);
    }
    
    /**
     * 
     * Enable/disable the BonSAI configuration file chooser (disabled by default).
     *
     * @param enable <code>true</code> to enable, <code>false</code> otherwise.
     */
    public void setConfigChooserEnabled(boolean enable) {
        L_BONSAI_CONFIG_FILE.setEnabled(enable);
        tfBonsaiConfigFile.setEnabled(enable);
        bBrowseConfigFile.setEnabled(enable);
        bLoadConfigFile.setEnabled(enable);
    }
}
