/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.util;

import java.util.Vector;

import javax.vecmath.Matrix3f;
import javax.vecmath.Matrix4f;
import javax.vecmath.Vector3f;
import javax.vecmath.Vector4f;

/**
 * 
 * @author flier
 */
public class HelmertTransformer {

    // Setup your origin und target system
    // TODO: make this non static
    private float[][] originsystem;
    private float[][] targetsystem;
    // This Array contains the translation vectors at column 0,1,2
    private float[][] transformvals = new float[3][3];
    // The derived rotationmatrix based on the origin and targetsystem
    private float[][] rotationmatrix = new float[3][3];
    // The derived eigenvalues
    private float[][][] eigenvalues = new float[2][1][1];
    // Center of gravity originsystem
    private float[] centerofgravity_origin = new float[3];
    // Center of gravity targetsystem
    private float[] centerofgravity_target = new float[3];
    // The scalefactor between origin and targetsystem, should be linear :)
    private float scalefactor = 0.0f;
    // The transformation Matrix
    private float[][] transformationsmatrix = new float[4][4];
    // The transformation Matrix4f
    private Matrix4f transformationsmatrix4f = new Matrix4f();
    // Global scope buffer
    private float m1 = 0.0f;
    private float m2 = 0.0f;
    private float mg = 0.0f;
    private float[][] redstartg = new float[3][3];
    private float[][] redtargetg = new float[3][3];

    /**
     * Transformer()
     * 
     * @param args
     *            the command line arguments == none
     */
    public HelmertTransformer() {

        originsystem = new float[3][3];
        targetsystem = new float[3][3];

        // Fill origin and target system with values (hardcoded)
        float[] o0 = { 2 * 420f, 2 * 66f, 2 * 480f };
        float[] o1 = { 2 * 414f, 2 * 68f, 2 * 480f };
        float[] o2 = { 2 * 417f, 2 * 69f, 2 * 478f };
        float[] t0 = { 420f, 66f, 477f };
        float[] t1 = { 414f, 68f, 477f };
        float[] t2 = { 417f, 69f, 475f };

        originsystem[0] = o0;
        originsystem[1] = o1;
        originsystem[2] = o2;

        targetsystem[0] = t0;
        targetsystem[1] = t1;
        targetsystem[2] = t2;

        doHelmertTransformation(originsystem, targetsystem);

        // Run Tests
        RunTests();
    }

    /**
     * HelmertTransformer()
     * 
     * @param systemO
     *            origin system
     * @param systemT
     *            target system
     */
    public HelmertTransformer(Vector<Vector<Double>> systemO,
            Vector<Vector<Double>> systemT) {
        if (systemO.size() != systemT.size()) {
            System.out.println("OriginSystem.length != TargetSystem.length!");
        } else {
            originsystem = new float[systemO.size()][systemO.size()];

            float[] o0 =
                    { ((Double) systemO.get(0).get(0)).floatValue(),
                            ((Double) systemO.get(0).get(1)).floatValue(),
                            ((Double) systemO.get(0).get(2)).floatValue() };

            float[] o1 =
                    { ((Double) systemO.get(1).get(0)).floatValue(),
                            ((Double) systemO.get(1).get(1)).floatValue(),
                            ((Double) systemO.get(1).get(2)).floatValue() };

            float[] o2 =
                    { ((Double) systemO.get(2).get(0)).floatValue(),
                            ((Double) systemO.get(2).get(1)).floatValue(),
                            ((Double) systemO.get(2).get(2)).floatValue() };

       

            originsystem[0] = o0;
            originsystem[1] = o1;
            originsystem[2] = o2;

            targetsystem = new float[systemT.size()][systemT.size()];
        

            float[] t0 =
                    { ((Double) systemT.get(0).get(0)).floatValue(),
                            ((Double) systemT.get(0).get(1)).floatValue(),
                            ((Double) systemT.get(0).get(2)).floatValue() };

            float[] t1 =
                    { ((Double) systemT.get(1).get(0)).floatValue(),
                            ((Double) systemT.get(1).get(1)).floatValue(),
                            ((Double) systemT.get(1).get(2)).floatValue() };

            float[] t2 =
                    { ((Double) systemT.get(2).get(0)).floatValue(),
                            ((Double) systemT.get(2).get(1)).floatValue(),
                            ((Double) systemT.get(2).get(2)).floatValue() };
            

            targetsystem[0] = t0;
            targetsystem[1] = t1;
            targetsystem[2] = t2;
            
            doHelmertTransformation(originsystem, targetsystem);
        }
    }

    /**
     * HelmertTransformer()
     * 
     * @param args
     *            invoke with given parameters: system lengths
     */

    public HelmertTransformer(float[][] systemO, float[][] systemT) {

        if (systemO.length != systemT.length) {
            System.out.println("OriginSystem.length != TargetSystem.length!");

        } else {
            originsystem = new float[systemO.length][systemO.length];
            for (int i = 0; i < systemO.length; i++) {
                originsystem[i] = systemO[i];
            }
            targetsystem = new float[systemT.length][systemT.length];
            for (int i = 0; i < systemT.length; i++) {
                targetsystem[i] = systemT[i];
            }
        }

        doHelmertTransformation(originsystem, targetsystem);

        // Run Tests
        RunTests();

    }

    /**
     * createNullMatrix()
     * 
     * @param <int> r == rows
     * @param <int> c == columns
     * @return <Array> Matrix mat
     * @description Contructs a matrix filled with zeros (float)
     * 
     **/
    public float[][] createNullMatrix(int r, int c) {

        float[][] mat = new float[r][c];
        for (int i = 0; i < r; i++) {
            for (int j = 0; j < c; j++) {
                mat[i][j] = 0.0f;
            }
        }
        return mat;
    }

    /**
     * transposeMatrix()
     * 
     * @param <Array>
     *            A
     * @return <Array> AT
     * @description Transposes the Matrix A and returns AT
     * 
     * 
     **/
    public float[][] transposeMatrix(float[][] mat) {
        float[][] transmat = createNullMatrix(mat[0].length, mat.length);
        for (int i = 0; i < mat.length; i++) {
            for (int j = 0; j < mat[i].length; j++) {
                transmat[j][i] = mat[i][j];
            }
        }

        return transmat;
    }

    /**
     * multiplyMatrix()
     * 
     * @param <Array>
     *            mat1
     * @param <Array>
     *            mat2
     * @return <Array> sum = mat1*mat2
     * @description Multiplies mat1(o,p), mat2(l,k) and return the sum(p,l)
     * 
     **/
    public float[][] multiplyMatrix(float[][] mat1, float[][] mat2) {

        if (mat1[0].length == mat2.length) {
            float[][] sum = createNullMatrix(mat1.length, mat2[0].length);
            for (int i = 0; i < mat1.length; i++) {
                for (int j = 0; j < mat2[0].length; j++) {
                    sum[i][j] = 0.0f;
                    for (int k = 0; k < mat1[0].length; k++) {
                        sum[i][j] += mat1[i][k] * mat2[k][j];
                    }
                }
            }
            return sum;
        }
        return null;
    }

    /*
     * deriveEigenwert()
     * 
     * @param <Array> mat1
     * 
     * @return <Array,<Array><Array>>
     * 
     * @description Derives the eigenvalues und eigenvectors - [Jaeger et al,
     * 2005] returns ZERO if the matrix is not quadratic or max iteration (20)
     * has been reached
     */
    public float[][][] deriveEigenwert(float[][] mat) {

        if (mat[0].length != mat.length) {
            return null;
        }
        int n = mat.length;
        float[][] eigen = mat;
        float[][] eigenvektoren = new float[n][n];
        for (int k = 0; k < n; k++) {
            for (int l = 0; l < n; l++) {
                eigenvektoren[k][l] = 0.0f;
            }
            eigenvektoren[k][k] = 1.0f;
        }

        float phi = Float.MAX_VALUE;
        int iter = 20;
        while (Math.abs(phi) > 1.0E-8 && iter-- > 0) {
            for (int i = 0; i < n - 1; i++) {
                for (int j = i + 1; j < n; j++) {
                    float[][] tmp = new float[n][n];
                    for (int k = 0; k < n; k++) {
                        for (int l = 0; l < n; l++) {
                            tmp[k][l] = eigenvektoren[k][l];
                        }
                    }
                    phi =
                            (float) (0.5f * Math.atan2(2.0 * eigen[i][j],
                                    eigen[i][i] - eigen[j][j]));

                    for (int k = 0; k < n; k++) {
                        for (int l = 0; l < n; l++) {
                            eigenvektoren[k][l] = 0.0f;
                        }
                        eigenvektoren[k][k] = 1.0f;
                    }

                    eigenvektoren[i][i] = (float) Math.cos(phi);
                    eigenvektoren[i][j] = (float) -Math.sin(phi);
                    eigenvektoren[j][i] = (float) Math.sin(phi);
                    eigenvektoren[j][j] = (float) Math.cos(phi);

                    float[][] eigentranspose = transposeMatrix(eigenvektoren);

                    eigen =
                            multiplyMatrix(eigentranspose, (multiplyMatrix(
                                    eigen, eigenvektoren)));

                    eigenvektoren = multiplyMatrix(tmp, eigenvektoren);
                }
            }
        }
        if (iter <= 0) {
            return null;
        }
        for (int i = 0; i < eigen.length; i++) {
            for (int j = 0; j < eigen[i].length; j++) {
                if (i != j) {
                    eigen[i][j] = 0.0f;
                }
            }
        }

        float[][][] result = new float[2][1][1];
        result[0] = eigenvektoren;
        result[1] = eigen;

        this.eigenvalues = result;

        return result;
    }

    /*
     * deriveCenterOfGravity()
     * 
     * @param <Array> mat1
     * 
     * @return <Array> result
     * 
     * @description Derives the Center of Gravity of a given pointcloud
     */
    public static float[] deriveCenterOfGravity(float[][] system) {
        float x, y, z;
        x = 0.0f;
        y = 0.0f;
        z = 0.0f;

        for (int i = 0; i < system.length; i++) {
            x += system[i][0];
            y += system[i][1];
            z += system[i][2];
        }

        float[] result = new float[3];
        result[0] = x / system.length;
        result[1] = y / system.length;
        result[2] = z / system.length;

        return result;
    }

    /*
     * doHelmertTransformation()
     * 
     * @param <Array> system
     * 
     * @return boolean
     * 
     * @description Derives the RationMatrix, Translationvectors and a
     * scalefacor, results are copied into the global scope
     */
    public boolean doHelmertTransformation(float[][] originsystem,
            float[][] targetsystem) {

        System.out.println("Deriving Transformation...");
        final long startZeit_1CPU = System.currentTimeMillis();

        if (originsystem.length > 0
                && originsystem.length == targetsystem.length) {

            float Sxx = 0.0f;
            float Sxy = 0.0f;
            float Sxz = 0.0f;
            float Syx = 0.0f;
            float Syy = 0.0f;
            float Syz = 0.0f;
            float Szx = 0.0f;
            float Szy = 0.0f;
            float Szz = 0.0f;

            float[] cs = deriveCenterOfGravity(originsystem);
            float[] ct = deriveCenterOfGravity(targetsystem);

            this.centerofgravity_origin = cs;
            this.centerofgravity_target = ct;

            float[][] redStart =
                    new float[originsystem.length][originsystem.length];
            float[][] redTarget =
                    new float[originsystem.length][originsystem.length];

            float N[][] = new float[4][4];

            float[][] t = new float[3][3];

            for (int i = 0; i < originsystem.length; i++) { // for1

                float[] bufforigin =
                        { originsystem[i][0] - cs[0],
                                originsystem[i][1] - cs[1],
                                originsystem[i][2] - cs[2] };
                float[] bufftarget =
                        { targetsystem[i][0] - ct[0],
                                targetsystem[i][1] - ct[1],
                                targetsystem[i][2] - ct[2] };

                redStart[i] = bufforigin;
                redTarget[i] = bufftarget;

                Sxx +=
                        ((originsystem[i][0] - cs[0]) * (targetsystem[i][0] - ct[0]));
                Sxy +=
                        ((originsystem[i][0] - cs[0]) * (targetsystem[i][1] - ct[1]));
                Sxz +=
                        ((originsystem[i][0] - cs[0]) * (targetsystem[i][2] - ct[2]));

                Syx +=
                        ((originsystem[i][1] - cs[1]) * (targetsystem[i][0] - ct[0]));
                Syy +=
                        ((originsystem[i][1] - cs[1]) * (targetsystem[i][1] - ct[1]));
                Syz +=
                        ((originsystem[i][1] - cs[1]) * (targetsystem[i][2] - ct[2]));

                Szx +=
                        ((originsystem[i][2] - cs[2]) * (targetsystem[i][0] - ct[0]));
                Szy +=
                        ((originsystem[i][2] - cs[2]) * (targetsystem[i][1] - ct[1]));
                Szz +=
                        ((originsystem[i][2] - cs[2]) * (targetsystem[i][2] - ct[2]));

            }

            this.redstartg = redStart;
            this.redtargetg = redTarget;

            float sx0, sx1, sx2, sx3, sy0, sy1, sy2, sy3, sz0, sz1, sz2, sz3, sw0, sw1, sw2, sw3;

            sx0 = Sxx + Syy + Szz;
            sx1 = Syz - Szy;
            sx2 = Szx - Sxz;
            sx3 = Sxy - Syx;

            sy0 = Syz - Szy;
            sy1 = Sxx - Syy - Szz;
            sy2 = Sxy + Syx;
            sy3 = Szx + Sxz;

            sz0 = Szx - Sxz;
            sz1 = Sxy + Syx;
            sz2 = -Sxx + Syy - Szz;
            sz3 = Syz + Szy;

            sw0 = Sxy - Syx;
            sw1 = Szx + Sxz;
            sw2 = Syz + Szy;
            sw3 = -Sxx - Syy + Szz;

            float[] buffn0 = { sx0, sx1, sx2, sx3 };
            float[] buffn1 = { sy0, sy1, sy2, sy3 };
            float[] buffn2 = { sz0, sz1, sz2, sz3 };
            float[] buffn3 = { sw0, sw1, sw2, sw3 };

            N[0] = buffn0;
            N[1] = buffn1;
            N[2] = buffn2;
            N[3] = buffn3;

            float[][][] VD = deriveEigenwert(N);
            if (VD == null) {
                System.out.println("VD == Null");
                return false;

            }

            int indexOfMaxEig = 0;

            float maxEig = (VD[1][indexOfMaxEig][indexOfMaxEig]);

            for (int i = 1; i < VD[1].length; i++) {
                if (maxEig < VD[1][i][i]) {
                    indexOfMaxEig = i;
                    maxEig = VD[1][i][i];
                }
            }

            float[] q =
                    { VD[0][0][indexOfMaxEig], VD[0][1][indexOfMaxEig],
                            VD[0][2][indexOfMaxEig], VD[0][3][indexOfMaxEig] };

            float[][] R = new float[3][3];

            float[] buffR0 =
                    { 2.0f * q[0] * q[0] - 1.0f + 2.0f * q[1] * q[1],
                            2.0f * (q[1] * q[2] - q[0] * q[3]),
                            2.0f * (q[1] * q[3] + q[0] * q[2]) };
            R[0] = buffR0;

            float[] buffR1 =
                    { 2.0f * (q[1] * q[2] + q[0] * q[3]),
                            2.0f * q[0] * q[0] - 1.0f + 2.0f * q[2] * q[2],
                            2.0f * (q[2] * q[3] - q[0] * q[1]) };
            R[1] = buffR1;

            float[] buffR2 =
                    { 2.0f * (q[1] * q[3] - q[0] * q[2]),
                            2.0f * (q[2] * q[3] + q[0] * q[1]),
                            2.0f * q[0] * q[0] - 1.0f + 2.0f * q[3] * q[3] };
            R[2] = buffR2;

            this.rotationmatrix = R;

            float[][] ps = new float[originsystem.length][originsystem.length];
            float[][] pt = new float[originsystem.length][originsystem.length];

            for (int i = 0; i < originsystem.length; i++) {
                ps[0] = this.redstartg[i];
                pt[0] = this.redtargetg[i];
                m1 +=
                        multiplyMatrix(pt, multiplyMatrix(R,
                                transposeMatrix(ps)))[0][0];
                m2 += multiplyMatrix(ps, transposeMatrix(ps))[0][0];
            }

            float m = (m1 / m2); // Math.abs(m1/m2);
            this.mg = m;
            this.scalefactor = Math.abs(m - 1.0f);

            float[][] tmp2 = new float[1][1];
            tmp2[0] = cs;
            t = multiplyMatrix(rotationmatrix, transposeMatrix(tmp2));
            for (int ii = 0; ii < 3; ii++) {
                t[ii][0] = ct[ii] - m * t[ii][0];
            }
            this.transformvals = t;

            final long endZeit_1CPU = System.currentTimeMillis();
            final long dauer_1CPU = endZeit_1CPU - startZeit_1CPU;
            System.out.println("...done");
            System.out.println("Time needed on 1 CPU: " + dauer_1CPU + " ms");

            return true;

        } else {
            System.out
                    .println("OriginSystem is < 0 or OriginSystem.length != StartSystem.length");
            return false;
        }

    }

    public Matrix3f array2RotationMatrix3f() {
        Matrix3f rotationmatrix3f =
                new Matrix3f(rotationmatrix[0][0], rotationmatrix[0][1],
                        rotationmatrix[0][2], rotationmatrix[1][0],
                        rotationmatrix[1][1], rotationmatrix[1][2],
                        rotationmatrix[2][0], rotationmatrix[2][1],
                        rotationmatrix[2][2]);
        return rotationmatrix3f;
    }

    public Vector3f array2TranslationVecMatrix3f() {
        Vector3f translationvector3f =
                new Vector3f(transformvals[0][0], transformvals[1][0],
                        transformvals[2][0]);
        return translationvector3f;
    }

    public float[][] getTransformationMatrix() {
        float[][] transformationsmatrixtmp = createNullMatrix(4, 4);
        for (int i = 0; i < 3; i++) {
            for (int j = 0; j < 3; j++) {
                transformationsmatrixtmp[i][j] = rotationmatrix[i][j];
            }
        }
        transformationsmatrixtmp[0][3] = transformvals[0][0];
        transformationsmatrixtmp[1][3] = transformvals[1][0];
        transformationsmatrixtmp[2][3] = transformvals[2][0];
        transformationsmatrixtmp[3][3] = 1.0f;

        transformationsmatrix = transformationsmatrixtmp;

        return transformationsmatrixtmp;
    }

    public Matrix4f getTransformationMatrix4f() {
        transformationsmatrix4f =
                new Matrix4f(rotationmatrix[0][0], rotationmatrix[0][1],
                        rotationmatrix[0][2], transformvals[0][0],
                        rotationmatrix[1][0], rotationmatrix[1][1],
                        rotationmatrix[1][2], transformvals[1][0],
                        rotationmatrix[2][0], rotationmatrix[2][1],
                        rotationmatrix[2][2], transformvals[2][0], 0.0f, 0.0f,
                        0.0f, 1.0f);
        return transformationsmatrix4f;
    }

    /* EXPERIMENTAL USE WITH CAUTION */
    public Vector4f MatrixtoAxisAngle(float[][] m) {

        float angle, x, y, z; // variables for result
        float epsilon = 0.01f; // margin to allow for rounding errors
        float epsilon2 = 0.1f; // margin to distinguish between 0 and 180
        // degrees

        if ((Math.abs(m[0][1] - m[1][0]) < epsilon)
                && (Math.abs(m[0][2] - m[2][0]) < epsilon)
                && (Math.abs(m[1][2] - m[2][1]) < epsilon)) {
            // singularity found
            // first check for identity matrix which must have +1 for all terms
            // in leading diagonaland zero in other terms
            if ((Math.abs(m[0][1] + m[1][0]) < epsilon2)
                    && (Math.abs(m[0][2] + m[2][0]) < epsilon2)
                    && (Math.abs(m[1][2] + m[2][1]) < epsilon2)
                    && (Math.abs(m[0][0] + m[1][1] + m[2][2] - 3) < epsilon2)) {
                // this singularity is identity matrix so angle = 0
                return new Vector4f(0, 1, 0, 0); // zero angle, arbitrary axis
            }
            // otherwise this singularity is angle = 180
            angle = (float) Math.PI;
            float xx = (m[0][0] + 1) / 2;
            float yy = (m[1][1] + 1) / 2;
            float zz = (m[2][2] + 1) / 2;
            float xy = (m[0][1] + m[1][0]) / 4;
            float xz = (m[0][2] + m[2][0]) / 4;
            float yz = (m[1][2] + m[2][1]) / 4;
            if ((xx > yy) && (xx > zz)) { // m[0][0] is the largest diagonal
                // term
                if (xx < epsilon) {
                    x = 0.0f;
                    y = 0.7071f;
                    z = 0.7071f;
                } else {
                    x = (float) Math.sqrt(xx);
                    y = xy / x;
                    z = xz / x;
                }
            } else if (yy > zz) { // m[1][1] is the largest diagonal term
                if (yy < epsilon) {
                    x = 0.7071f;
                    y = 0.0f;
                    z = 0.7071f;
                } else {
                    y = (float) Math.sqrt(yy);
                    x = xy / y;
                    z = yz / y;
                }
            } else { // m[2][2] is the largest diagonal term so base result on
                // this
                if (zz < epsilon) {
                    x = 0.7071f;
                    y = 0.7071f;
                    z = 0.0f;
                } else {
                    z = (float) Math.sqrt(zz);
                    x = xz / z;
                    y = yz / z;
                }
            }
            return new Vector4f(angle, x, y, z); // return 180 deg rotation
        }
        // as we have reached here there are no singularities so we can handle
        // normally
        float s =
                (float) Math.sqrt((m[2][1] - m[1][2]) * (m[2][1] - m[1][2])
                        + (m[0][2] - m[2][0]) * (m[0][2] - m[2][0])
                        + (m[1][0] - m[0][1]) * (m[1][0] - m[0][1])); // used to
        // normalise
        if (Math.abs(s) < 0.001) {
            s = 1;
        }
        // prevent divide by zero, should not happen if matrix is orthogonal and
        // should be
        // caught by singularity test above, but I've left it in just in case
        angle = (float) Math.acos((m[0][0] + m[1][1] + m[2][2] - 1) / 2);
        x = (m[2][1] - m[1][2]) / s;
        y = (m[0][2] - m[2][0]) / s;
        z = (m[1][0] - m[0][1]) / s;

        return new Vector4f(angle, x, y, z);
    }

    public float getScalefactor() {
        return scalefactor;
    }

    /*
     * RunTests()
     * 
     * @description Run various tests concerning the transformation
     */
    public void RunTests() {

        System.out.println("---------\n---------");

        for (int i = 0; i < centerofgravity_origin.length; i++) {
            System.out
                    .println("CoG Originsystem: " + centerofgravity_origin[i]);
        }

        System.out.println("---------\n---------");

        for (int i = 0; i < centerofgravity_target.length; i++) {
            System.out
                    .println("CoG Targetsystem: " + centerofgravity_target[i]);
        }

        System.out.println("---------\n---------");

        for (int i = 0; i < eigenvalues[0][1].length; i++) {
            for (int j = 0; j < eigenvalues[0][1].length; j++) {
                System.out.println("Eigenvektoren: " + eigenvalues[0][i][j]);
            }
        }

        System.out.println("---------\n---------");

        for (int i = 0; i < eigenvalues[1][1].length; i++) {
            for (int j = 0; j < eigenvalues[1][1].length; j++) {
                System.out.println("Eigenwerte: " + eigenvalues[1][i][j]);
            }
        }

        System.out.println("---------\n---------");

        for (int ii = 0; ii < rotationmatrix.length; ii++) {
            for (int iii = 0; iii < rotationmatrix.length; iii++) {
                System.out.println("3x3 RotMatrix: " + rotationmatrix[ii][iii]
                        + " " + "Zeile " + ii + " Spalte " + iii);
            }
        }

        System.out.println("---------\n---------");
        // System.out.println("M1: " + m1);
        // System.out.println("M2: " + m2);
        // System.out.println("M1/M2: " + mg);
        System.out.println("Scalefactor: " + scalefactor);

        System.out.println("---------\n---------");

        for (int i = 0; i < transformvals.length; i++) {
            System.out.println("|Transformation|: " + transformvals[i][0]);
        }

        System.out.println("---------\n---------");

        System.out.println("RotationMatrix3f: " + array2RotationMatrix3f());

        System.out.println("---------\n---------");

        System.out.println("TranslationVector3f: "
                + array2TranslationVecMatrix3f());

        System.out.println("---------\n---------");

        System.out.println("AxisAngles: " + MatrixtoAxisAngle(rotationmatrix));

        System.out.println("---------\n---------");

        System.out.println("Transformationmatrix:");

        getTransformationMatrix();
        for (int i = 0; i < transformationsmatrix.length; i++) {
            for (int j = 0; j < transformationsmatrix.length; j++) {
                System.out.println("4x4TransformationMatrix: "
                        + transformationsmatrix[i][j] + " " + "Zeile " + i
                        + " Spalte " + j);
            }
        }

        System.out.println("---------\n---------");

        System.out.println("Transformationmatrix4f: " + "\n"
                + getTransformationMatrix4f());

    }
}
