/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.util.helper;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.core.SensorListener;
import de.unibi.airobots.bonsai.util.helper.GoalReachedWaiter.RefinementStrategy.UnrefinableException;
import de.unibi.airobots.btl.data.navigation.GlobalPlannerStateChange;
import de.unibi.airobots.btl.data.navigation.GlobalPlannerStateChange.PlannerState;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData.GoalType;
import de.unibi.airobots.btl.data.navigation.PositionData;
import de.unibi.airobots.btl.data.navigation.SpeedData;
import de.unibi.airobots.btl.units.AngleUnit;
import de.unibi.airobots.btl.units.LengthUnit;
import de.unibi.airobots.btl.units.SpeedUnit;

import java.io.IOException;
import java.util.Collection;
import java.util.List;
import org.apache.log4j.Logger;

/**
 * A class that helps waiting to reach a navigation goal. To determined if a
 * goal can't be reached a heuristic is used.
 *
 * @author jwienke
 */
public class GoalReachedWaiter implements
        SensorListener<GlobalPlannerStateChange> {

    /**
     * Results types after waiting for a goal.
     *
     * @author jwienke
     */
    public enum ReachStatus {

        /**
         * Goal was reached successfully.
         */
        REACHED,
        /**
         * Goal was not reached because of a timeout.
         */
        NOT_REACHED_TIMEOUT,
        /**
         * Goal was not reached because the heuristic decided that it could not
         * be reached.
         */
        NOT_REACHED_HEURISTIC,
        /**
         * Goal was not reached because the global planner could not find a path
         * to the given global goal.
         */
        NOT_REACHED_BLOCKED,
        /**
         * Goal was not reached because waiting for it was interrupted.
         */
        NOT_REACHED_INTERRUPTED,
        /**
         * Waiting to reach the goal was canceled by the custom predicate.
         */
        CANCELED_BY_PREDICATE,
        /**
         * Waiting to reach the goal was canceled because the refinement
         * strategy threw an error.
         */
        CANCELED_BY_REFINEMENT
    }

    /**
     * A custom predicate that can be used to cancel waiting externally.
     *
     * @author jwienke
     */
    public interface CancelWaitingPredicate {

        /**
         * Predicate function.
         * 
         * @return <code>true</code> to cancel waiting for reaching a goal, else
         *         <code>false</code>
         */
        boolean cancelWaiting();
    }

    /**
     * A strategy that refines the goal while reaching it.
     *
     * @author jwienke
     */
    public interface RefinementStrategy {

        /**
         * Exception that may be thrown by a {@link RefinementStrategy} if the
         * goal cannot be refined and it is not desirable to wait for reaching
         * it.
         *
         * @author jwienke
         */
        public class UnrefinableException extends Exception {

            /**
             * Serial version UID.
             */
            private static final long serialVersionUID = 4230547229435217077L;

            /**
             * Constructor.
             *
             * @param message
             *            reason why the goal could not be refined.
             */
            public UnrefinableException(String message) {
                super(message);
            }

            /**
             * Constructor.
             *
             * @param message
             *            reason why the goal could not be refined.
             * @param cause
             *            root cause
             */
            public UnrefinableException(String message, Throwable cause) {
                super(message, cause);
            }
        }

        /**
         * Returns a refined goal.
         *
         * @return new goal to approach
         * @throws UnrefinableException
         *             if the goal cannot be refined and it is not desirable
         *             anymore to wait for reaching it
         */
        NavigationGoalData refineGoal() throws UnrefinableException;
    }

    private class DefaultRefinementStrategy implements RefinementStrategy {

        private NavigationGoalData constantGoal;

        public DefaultRefinementStrategy(final NavigationGoalData constantGoal) {
            this.constantGoal = constantGoal;
        }

        @Override
        public NavigationGoalData refineGoal() {
            return constantGoal;
        }
    }

    private class DefaultPredicate implements CancelWaitingPredicate {

        @Override
        public boolean cancelWaiting() {
            return false;
        }
    }
    private Logger logger = Logger.getLogger(getClass());
    /**
     * Default timeout used for waiting on reaching a goal.
     */
    public static final long DEFAULT_TIMEOUT = 20000;
    /**
     * Default time in ms to wait between checks for reaching goal.
     */
    public static final long DEFAULT_WAIT_TIME = 600;
    /**
     * Default number of measurements to collect before using the heuristic to
     * determine if a goal can't be reached. This is also the absolute number of
     * measurements that were taken into account.
     */
    public static final int DEFAULT_STEPS = 20;
    private final Sensor<PositionData> positionSensor;
    private final Sensor<SpeedData> speedSensor;
    private final NavigationActuator navigationActuator;
    private final long waitTime;
    private final int steps;
    private PlannerState currentGlobalPlanState;
    private volatile boolean receivedFirstChangeToPlanningState = false;

    /**
     * Creates a helper class to wait on navigation goals using custom values.
     * 
     * @param positionSensor
     *            The sensor used to get the position.
     * @param speedSensor
     *            The sensor used to get the current speed.
     * @param waitTime
     *            time to wait between checks
     * @param steps
     *            number of measurements to take into account
     */
    public GoalReachedWaiter(Sensor<PositionData> positionSensor,
            Sensor<SpeedData> speedSensor,
            Sensor<GlobalPlannerStateChange> gplanChangeSensor,
            NavigationActuator navigationActuator, long waitTime, int steps) {
        this.positionSensor = positionSensor;
        this.speedSensor = speedSensor;
        this.waitTime = waitTime;
        this.steps = steps;
        this.navigationActuator = navigationActuator;
        gplanChangeSensor.addSensorListener(this);

        // TODO was war das nochmal???????
        // if (navigationActuator instanceof DirectNavigationActuator) {
        // DirectNavigationActuator actuator = (DirectNavigationActuator)
        // navigationActuator;
        // // currentGlobalPlanState = actuator.getCurrentPlannerState();
        // actuator.addPlanStateListener(this);
        // }
    }

    /**
     * Creates a helper class to wait on navigation goals using default values.
     *
     * @param positionSensor
     *            The sensor used to get the position.
     * @param speedSensor
     *            The sensor used to get the speeds.
     * @param navigationActuator
     *            The actuator used to check if we have reached the goal.
     */
    public GoalReachedWaiter(Sensor<PositionData> positionSensor,
            Sensor<SpeedData> speedSensor,
            Sensor<GlobalPlannerStateChange> gplanChangeSensor,
            NavigationActuator navigationActuator) {
        this(positionSensor, speedSensor, gplanChangeSensor, navigationActuator,
                DEFAULT_WAIT_TIME, DEFAULT_STEPS);
    }

    private Sensor<PositionData> getPositionSensor() {
        return positionSensor;
    }

    private Sensor<SpeedData> getSpeedSensor() {
        return speedSensor;
    }

    /**
     * Waits for the robot to reach the navigation goal. If the robot cannot
     * reach the goal this method also stops. The return value indicates whether
     * the goal was reached successfully or not. This method blocks for a
     * possibly long time.
     *
     * This method reacts on {@link InterruptedException} by returning
     * <code>false</code>.
     *
     * @param goal
     *            the goal to reach
     * @param timeout
     *            timeout after which this waiter always reports that the goal
     *            could not be reached
     * @param predicate
     *            a custom predicate that cancels waiting on the goal. If the
     *            predicate canceled waiting, <code>true</code> will be
     *            returned.
     * @param refinementStrategy
     *            a strategy to refine goals while moving
     * @return status for reaching the goal
     * @throws IOException
     *             error accessing one of the sensors
     */
    public ReachStatus waitForReachingGoal(final NavigationGoalData goal,
            final long timeout, CancelWaitingPredicate predicate,
            RefinementStrategy refinementStrategy) throws IOException {

        logger.debug("waiting to reach goal " + goal.toDocument(NavigationGoalData.class).toXML());

        long startTime = System.currentTimeMillis();
        NavigationGoalData currentGoal = goal;

        try {

            ConstantSizeQueue<PositionData> lastPositions = new ConstantSizeQueue<PositionData>(
                    steps);
            ConstantSizeQueue<SpeedData> lastSpeeds = new ConstantSizeQueue<SpeedData>(
                    steps);

            boolean firstRun = true;

            while (true) {

                //logger.trace("new waiting loop");

                // If we are waiting for a global goal...
                if (goal.getType() == GoalType.LOCAL
                        || receivedFirstChangeToPlanningState) {

                    if (getCurrentGlobalPlannerState() == PlannerState.PATH_BLOCKED) {
                        logger.info("Path was blocked!");
                        return ReachStatus.NOT_REACHED_BLOCKED;
                    }

                    if (firstRun) {
                        if (goal.getType() == GoalType.GLOBAL) {
                            Thread.sleep(1000);
                        }
                        firstRun = false;
                    }

                    // stop waiting if the robot reached the (right) goal
                    boolean goalReached = navigationActuator.isGoalReached().isStopped();
                    if (goalReached) {
                        logger.info("Goal was reached.");
                        return ReachStatus.REACHED;
                    }

                    // stop waiting if the predicates forces this
                    if (predicate.cancelWaiting()) {
                        logger.info("Predicate canceled waiting.");
                        return ReachStatus.CANCELED_BY_PREDICATE;
                    }

                    // check timeout
                    if (System.currentTimeMillis() - startTime > timeout) {
                        logger.info("Reaching goal timed out.");
                        return ReachStatus.NOT_REACHED_TIMEOUT;
                    }
                }

                Thread.sleep(waitTime);

                // collect data
                //logger.trace("begin reading new data");
                lastPositions.add(getPositionSensor().readLast());
                //logger.trace("got position data");
                lastSpeeds.add(getSpeedSensor().readLast());
                //logger.trace("end reading new data");

                // refine goal
                NavigationGoalData refinedGoal = refinementStrategy.refineGoal();
                if (!isSameGoal(currentGoal, refinedGoal)) {
                    // if we got a new goal, we need to reset the start time for
                    // the timeout
                    //logger.trace("Resetting timeout start.");
                    startTime = System.currentTimeMillis();
                    navigationActuator.setGoal(refinedGoal);
                    currentGoal = refinedGoal;
                } else {
                    //logger.trace("Not resetting timout start.");
                }

                // don't use heuristic if there is not enough data collected
                if (lastPositions.size() != steps && lastSpeeds.size() != steps) {
                    //logger.trace("not enough data, skiping heurisitc");
                    continue;
                }

                // use heuristic to determine if robot can't reach goal
                boolean stillMoving = isStillMoving(lastSpeeds);
                boolean distanceDecreasing = isDistanceDecreasing(goal,
                        lastPositions);

                logger.trace("Waiting for reaching goal, stillMoving = "
                        + stillMoving + ", distanceDecreasing = "
                        + distanceDecreasing);

                if (!stillMoving && !distanceDecreasing) {
                    logger.info("Heuristic determined that the goal "
                            + "cannot be reached (stillMoving = " + stillMoving
                            + ", distanceDecreasing = " + distanceDecreasing
                            + ").");
                    return ReachStatus.NOT_REACHED_HEURISTIC;
                }

            }

        } catch (InterruptedException e) {
            logger.info("Interrupted while waiting for "
                    + "robot reaching goal.", e);
            return ReachStatus.NOT_REACHED_INTERRUPTED;
        } catch (UnrefinableException e) {
            logger.info("Refinement strategy cancelled waiting "
                    + "for reaching the goal.", e);
            return ReachStatus.CANCELED_BY_REFINEMENT;
        }

    }

    /**
     * Resets internal status variables. Must be used after setting global
     * navigation goals, if this waiter should be used more than once.
     */
    public void reset() {
        this.currentGlobalPlanState = null;
        this.receivedFirstChangeToPlanningState = false;
    }

    private boolean isSameGoal(NavigationGoalData currentGoal,
            NavigationGoalData refinedGoal) {

        // TODO what about yaw?

        double dist = Math.sqrt(Math.pow(currentGoal.getX(LengthUnit.METER)
                - refinedGoal.getX(LengthUnit.METER), 2)
                + Math.pow(currentGoal.getY(LengthUnit.METER) - refinedGoal.getY(LengthUnit.METER), 2));
        double minTolerance = Math.min(currentGoal.getCoordinateTolerance(LengthUnit.METER),
                refinedGoal.getCoordinateTolerance(LengthUnit.METER));

        //logger.trace("Comparing goals:\n" + currentGoal + "\n" + refinedGoal
        //		+ "\n" + "dist = " + dist + ", minTolerance = " + minTolerance);

        return (dist < minTolerance);

    }

    /**
     * Tells if the distance to a goal is decreasing.
     *
     * @param goal
     *            goal to reach
     * @param lastPositions
     *            list of last positions of the robot. First element must be the
     *            oldest
     * @return <code>true</code> if the distance is decreasing ,else
     *         <code>false</code>
     */
    private boolean isDistanceDecreasing(PositionData goal,
            List<PositionData> lastPositions) {

        int half = lastPositions.size() / 2;
        double firstDistanceRaw = 0.0;
        for (int i = 0; i < half; i++) {
            firstDistanceRaw += getDistance(lastPositions.get(0), goal);
        }
        double firstDistance = firstDistanceRaw / half;

        double secondDistanceRaw = 0.0;
        for (int i = half; i < lastPositions.size(); i++) {
            secondDistanceRaw += getDistance(lastPositions.get(i), goal);
        }
        double secondDistance = secondDistanceRaw
                / (lastPositions.size() - half);

        return secondDistance < firstDistance;

    }

    private double getDistance(PositionData d1, PositionData d2) {
        return Math.sqrt(Math.pow(d1.getX(LengthUnit.METER) - d2.getX(LengthUnit.METER), 2)
                + Math.pow(d1.getY(LengthUnit.METER) - d2.getY(LengthUnit.METER), 2));
    }

    /**
     * Tells if the robot is still moving.
     *
     * @param lastSpeeds
     *            collection of last speed measurements
     * @return <code>true</code> if robot is still moving, else
     *         <code>false</code>
     */
    private boolean isStillMoving(Collection<SpeedData> lastSpeeds) {

        final double NO_TRANSLATIONAL_SPEED = 0.00001;
        final double NO_ROTATORY_SPEED = 0.00001;

        //logger.info("Checking still moving");
        boolean wasMoving = false;
        for (SpeedData d : lastSpeeds) {
            //logger.info("Moving data: " + d);
            wasMoving |= (d.getSpeedRotation(AngleUnit.RADIAN) > NO_ROTATORY_SPEED)
                    || (d.getSpeedTranslation(SpeedUnit.METER_PER_SEC) > NO_TRANSLATIONAL_SPEED);
        }

        return wasMoving;
    }

    /**
     * Waits for the robot to reach the navigation goal. If the robot cannot
     * reach the goal this method also stops. The return value indicates whether
     * the goal was reached successfully or not. This method blocks for a
     * possibly long time and uses {@link #DEFAULT_TIMEOUT}.
     *
     * This method reacts on {@link InterruptedException} by returning
     * <code>false</code>.
     *
     * @param goal
     *            the goal to reach
     * @return status after waiting to reach the goal
     * @throws IOException
     *             error accessing one of the sensors
     */
    public ReachStatus waitForReachingGoal(NavigationGoalData goal)
            throws IOException {
        return waitForReachingGoal(goal, DEFAULT_TIMEOUT,
                new DefaultPredicate(), new DefaultRefinementStrategy(goal));
    }

    /**
     * Waits for the robot to reach the navigation goal. If the robot cannot
     * reach the goal this method also stops. The return value indicates whether
     * the goal was reached successfully or not. This method blocks for a
     * possibly long time and uses {@link #DEFAULT_TIMEOUT}.
     *
     * This method reacts on {@link InterruptedException} by returning
     * <code>false</code>.
     *
     * @param goal
     *            the goal to reach
     * @param predicate
     *            a custom predicate that cancels waiting on the goal. If the
     *            predicate canceled waiting, <code>true</code> will be
     *            returned.
     * @return status after waiting to each the goal
     * @throws IOException
     *             error accessing one of the sensors
     */
    public ReachStatus waitForReachingGoal(NavigationGoalData goal,
            CancelWaitingPredicate predicate) throws IOException {
        return waitForReachingGoal(goal, DEFAULT_TIMEOUT, predicate,
                new DefaultRefinementStrategy(goal));
    }

    /**
     * Waits for the robot to reach the navigation goal. If the robot cannot
     * reach the goal this method also stops. The return value indicates whether
     * the goal was reached successfully or not.
     *
     * This method reacts on {@link InterruptedException} by returning
     * <code>false</code>.
     *
     * @param goal
     *            the goal to reach
     * @param timeout
     *            timeout after which the goal will be declared as not reached
     * @return status after waiting to each the goal
     * @throws IOException
     *             error accessing one of the sensors
     */
    public ReachStatus waitForReachingGoal(NavigationGoalData goal, long timeout)
            throws IOException {
        return waitForReachingGoal(goal, timeout, new DefaultPredicate(),
                new DefaultRefinementStrategy(goal));
    }

    /**
     * Waits for the robot to reach the navigation goal. If the robot cannot
     * reach the goal this method also stops. The return value indicates whether
     * the goal was reached successfully or not.
     *
     * This method reacts on {@link InterruptedException} by returning
     * <code>false</code>.
     *
     * @param goal
     *            the goal to reach
     * @param timeout
     *            timeout after which the goal will be declared as not reached
     * @param refinementStrategy
     *            a strategy to refine goals while moving
     * @return status after waiting to each the goal
     * @throws IOException
     *             error accessing one of the sensors
     */
    public ReachStatus waitForReachingGoal(NavigationGoalData goal,
            long timeout, RefinementStrategy refinementStrategy)
            throws IOException {
        return waitForReachingGoal(goal, timeout, new DefaultPredicate(),
                refinementStrategy);
    }

    /**
     * Waits for the robot to reach the navigation goal. If the robot cannot
     * reach the goal this method also stops. The return value indicates whether
     * the goal was reached successfully or not. This method blocks for a
     * possibly long time.
     *
     * This method reacts on {@link InterruptedException} by returning
     * <code>false</code>.
     *
     * @param goal
     *            the goal to reach
     * @param refinementStrategy
     *            a strategy to refine goals while moving
     * @return status after waiting to each the goal
     * @throws IOException
     *             error accessing one of the sensors
     */
    public ReachStatus waitForReachingGoal(NavigationGoalData goal,
            RefinementStrategy refinementStrategy) throws IOException {
        return waitForReachingGoal(goal, DEFAULT_TIMEOUT,
                new DefaultPredicate(), refinementStrategy);
    }

    /**
     * Waits for the robot to reach the navigation goal. If the robot cannot
     * reach the goal this method also stops. The return value indicates whether
     * the goal was reached successfully or not. This method blocks for a
     * possibly long time.
     *
     * This method reacts on {@link InterruptedException} by returning
     * <code>false</code>.
     *
     * @param goal
     *            the goal to reach
     * @param timeout
     *            timeout after which this waiter always reports that the goal
     *            could not be reached
     * @param predicate
     *            a custom predicate that cancels waiting on the goal. If the
     *            predicate canceled waiting, <code>true</code> will be
     *            returned.
     * @return status for reaching the goal
     * @throws IOException
     *             error accessing one of the sensors
     */
    public ReachStatus waitForReachingGoal(NavigationGoalData goal,
            long timeout, CancelWaitingPredicate predicate) throws IOException {
        return waitForReachingGoal(goal, timeout, predicate,
                new DefaultRefinementStrategy(goal));
    }

    private synchronized PlannerState getCurrentGlobalPlannerState() {
        return this.currentGlobalPlanState;
    }

    /*
     * (non-Javadoc)
     *
     * @see de.unibi.airobots.btl.data.GlobalPlannerStateChange
     */
    @Override
    public void newDataAvailable(GlobalPlannerStateChange newData) {

        this.currentGlobalPlanState = newData.getNewState();

        if (!receivedFirstChangeToPlanningState
                && newData.getNewState() == PlannerState.PLANNING) {
            receivedFirstChangeToPlanningState = true;
        }

    }
}
