/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.util.slam.gui;

import static java.lang.Math.cos;
import static java.lang.Math.sin;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import javax.swing.JPanel;
import javax.swing.SwingUtilities;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.util.Pair;
import de.unibi.airobots.bonsai.util.helper.annotation.MapArea;
import de.unibi.airobots.bonsai.util.helper.annotation.MapArea.MapAreaType;
import de.unibi.airobots.bonsai.util.helper.annotation.MapPoint;
import de.unibi.airobots.bonsai.util.helper.annotation.MapRegion;
import de.unibi.airobots.bonsai.util.slam.SlamTools;
import de.unibi.airobots.btl.data.map.BinarySlamMap;
import de.unibi.airobots.btl.units.AngleUnit;
import de.unibi.airobots.btl.units.LengthUnit;

public class MapAnnotationPanel extends JPanel {

	/** To make Serializable happy... */
	private static final long serialVersionUID = 3347424196567633051L;

	/** For logging output. */
	private static final Logger LOGGER = Logger.getLogger(MapPanel.class);

	/** For getting the map. */
	private Sensor<BinarySlamMap> mapSensor;

	/** For notifying people of clicks into the world map. */
	private final List<WorldClickListener> listeners = new ArrayList<WorldClickListener>();

	/** The last image version of the slam map. */
	private BufferedImage slamImage;
	/** The latest slam data. */
	private BinarySlamMap slamMap;

	/** Used for periodically updating the map data. */
	private final ScheduledExecutorService executor = Executors
			.newSingleThreadScheduledExecutor();

	/** Used for transforming the image. */
	private AffineTransform imageTransform;
	/** Used for transforming all the things drawn on top of the image. */
	private AffineTransform pointTransform;

	private HashMap<String, MapArea> areas;

	private MapRegion currentRegion;

	// TODO: Rotate the map (world axis aligned to window)?
	// TODO: Find some elegant method for the zoom level?

	/**
	 * Creates a new map panel.
	 * 
	 * @param mapSensor
	 *            Used for getting the current SLAM map, should not be null.
	 * @param navigation
	 *            Used for getting the current goal of the navigation algorithm.
	 * @param positionSensor
	 *            Used for getting the current robot position.
	 * @param planSensor
	 *            Used for getting the current global plan.
	 * @param updateTime
	 *            How fast (in ms) should we try to update the SLAM map?
	 */
	public MapAnnotationPanel(final Sensor<BinarySlamMap> mapSensor,
			long updateTime) {
		super(new BorderLayout());

		this.setPreferredSize(new Dimension(570, 500));

		this.mapSensor = mapSensor;

		this.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(MouseEvent e) {
				notifyWorldClickListeners(e);
			}

		});

		executor.scheduleWithFixedDelay(new Runnable() {

			@Override
			public void run() {
				updateMap();
			}
		}, 200, updateTime, TimeUnit.MILLISECONDS);

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#finalize()
	 */
	@Override
	// CHECKSTYLE:OFF
	protected void finalize() throws Throwable {
		// CHECKSTYLE:ON
		super.finalize();
		executor.shutdownNow();
	}

	/**
	 * Tries to update the SLAM map from the sensor.
	 */
	private synchronized void updateMap() {
		try {
			slamMap = mapSensor.readLast(0);
			if (slamMap != null) {
				slamImage = SlamTools.slamMapToImage(slamMap);

				// TODO: Make scaling respect the maps aspect ratio?
				imageTransform = new AffineTransform(getWidth()
						/ (double) slamMap.getWidth(), 0, 0, getHeight()
						/ (double) slamMap.getHeight(), 0, 0);
				pointTransform = new AffineTransform(imageTransform);
				pointTransform.scale(1, -1);
				pointTransform.translate(0, -(slamMap.getHeight() - 1));
			}

			SwingUtilities.invokeLater(new Runnable() {

				@Override
				public void run() {
					repaint();
				}
			});
			// CHECKSTYLE:OFF we really want to catch everything here
		} catch (Throwable e) {
			// CHECKSTYLE:ON
			LOGGER.warn("Throwable caught while updating map panel.", e);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.swing.JComponent#paint(java.awt.Graphics)
	 */
	@Override
	public void paint(Graphics g) {

		Graphics2D g2 = (Graphics2D) g;

		g2.clearRect(0, 0, getWidth(), getHeight());

		drawMap(g2);
	}

	/**
	 * Draws the SLAM map.
	 * 
	 * @param g2
	 *            The {@link Graphics2D} object used to draw.
	 */
	private synchronized void drawMap(Graphics2D g2) {
		if (slamImage != null) {
			// Draw transformed image
			g2.drawImage(slamImage, imageTransform, this);

			// Draw red circle at current slam position
			int slamPosX = slamMap.x2idx(slamMap.getX(LengthUnit.METER),
					LengthUnit.METER);
			int slamPosY = slamMap.y2idx(slamMap.getY(LengthUnit.METER),
					LengthUnit.METER);
			g2.setColor(Color.YELLOW);
			double[] slamPos = new double[] { slamPosX, slamPosY };
			pointTransform.transform(slamPos, 0, slamPos, 0, 1);
			drawPose(g2, (int) slamPos[0], (int) slamPos[1],
					slamMap.getYaw(AngleUnit.RADIAN), true, String.format(
							"(%.3f,%.3f)", slamMap.getX(LengthUnit.METER),
							slamMap.getY(LengthUnit.METER)));

			if (areas != null) {
				for (MapArea mapArea : areas.values()) {

					for (MapRegion region : mapArea.getRegions()) {
						g2.setColor(getAreaColor(mapArea.getType()));
						drawRegion(g2, region);

						if (region.getViewPoint() != null) {

							g2.setColor(Color.pink);
							double[] point = new double[] {
									slamMap.x2idx(
											region.getViewPoint().getX(
													LengthUnit.METER),
											LengthUnit.METER),
									slamMap.y2idx(
											region.getViewPoint().getY(
													LengthUnit.METER),
											LengthUnit.METER) };
							pointTransform.transform(point, 0, point, 0, 1);
							int x = (int) point[0];
							int y = (int) point[1];
							drawPose(
									g2,
									x,
									y,
									region.getViewPoint().getYaw(
											AngleUnit.RADIAN), true, "");
						}

					}
				}
			}

			if (currentRegion != null) {
				g2.setColor(Color.blue);
				drawRegion(g2, currentRegion);

				if (currentRegion.getViewPoint() != null) {

					g2.setColor(Color.pink);
					double[] point = new double[] {
							slamMap.x2idx(
									currentRegion.getViewPoint().getX(
											LengthUnit.METER), LengthUnit.METER),
							slamMap.y2idx(
									currentRegion.getViewPoint().getY(
											LengthUnit.METER), LengthUnit.METER) };
					pointTransform.transform(point, 0, point, 0, 1);
					int x = (int) point[0];
					int y = (int) point[1];
					drawPose(
							g2,
							x,
							y,
							currentRegion.getViewPoint().getYaw(
									AngleUnit.RADIAN), true, "");
				}

			}

		}
	}

	private void drawRegion(Graphics2D g2, MapRegion region) {
		Polygon polygon = new Polygon();
		for (MapPoint p : region.getPoints()) {
			addPoint2Polygon(polygon, p.getX(), p.getY());
		}

		g2.drawPolygon(polygon);
		g2.drawString(region.getLabel(), (float) polygon.getBounds2D()
				.getCenterX(), (float) polygon.getBounds2D().getCenterY());
	}

	private Color getAreaColor(MapAreaType type) {

		switch (type) {
		case NOGO:
			return Color.RED;
		case INTEREST:
			return Color.GREEN;
		case PERSON:
			return Color.YELLOW;
		case ROOM:
			return Color.ORANGE;
		case DISTURB:
			return Color.MAGENTA;
		case NO_TYPE:
		default:
			return Color.CYAN;

		}

	}

	private void addPoint2Polygon(Polygon polygon, double dx, double dy) {
		double[] point = new double[] { slamMap.x2idx(dx, LengthUnit.METER),
				slamMap.y2idx(dy, LengthUnit.METER) };
		pointTransform.transform(point, 0, point, 0, 1);
		int x = (int) point[0];
		int y = (int) point[1];
		polygon.addPoint(x, y);
	}

	/**
	 * Draw a pose (position and optionally orientation) in the map.
	 * 
	 * @param g2
	 *            The graphics object used for drawing.
	 * @param x
	 *            The x coordinate for the pose (in image coordinates!).
	 * @param y
	 *            The y coordinate for the pose (in image coordinates!).
	 * @param yaw
	 *            The orientation angle (0.0 points to the right).
	 * @param drawOrientation
	 *            If we should draw the orientation.
	 * @param description
	 *            Optional description displayed at the pose.
	 */
	private void drawPose(Graphics2D g2, int x, int y, double yaw,
			boolean drawOrientation, String description) {
		// Draw a circle centered on the position
		int circleRadius = 5;
		g2.drawOval(x - circleRadius, y - circleRadius, (circleRadius * 2) + 1,
				(circleRadius * 2) + 1);

		// Optionally draw a line from the position in the direction of the
		// orientation
		if (drawOrientation) {
			float lineLength = circleRadius + 3;
			g2.drawLine(x, y, x + (int) (cos(yaw) * lineLength), y
					- (int) (sin(yaw) * lineLength));
		}

		// Optionally draw a description to the top right of the circle
		if (!(description == null || "".equals(description.trim()))) {
			g2.drawString(description, x + circleRadius, y - circleRadius);
		}
	}

	/**
	 * Converts pixel coordinates in this panel to coordinates in the slam maps
	 * occupancy grid.
	 * 
	 * @param panelCoordinates
	 *            Pixel coordinates in this panel, must not be null.
	 * @return Grid indices in the occupancy grid.
	 */
	private Pair<Integer> panelCoordinates2Grid(Pair<Integer> panelCoordinates) {
		try {
			// We just reverse the transformation from grid into pixel
			// coordinates and use that to transform the coordinates in the
			// other direction.
			AffineTransform reverse = pointTransform.createInverse();
			double[] point = new double[] { panelCoordinates.getFirst(),
					panelCoordinates.getSecond() };
			reverse.transform(point, 0, point, 0, 1);
			return new Pair<Integer>((int) point[0], (int) point[1]);
		} catch (NoninvertibleTransformException e) {
			// This should normally never happen!
			LOGGER.warn("Could not invert SLAM map transform.", e);
			throw new IllegalStateException("Could not invert transformation "
					+ "that should be invertible?!");
		}
	}

	/**
	 * Notifies all listeners of a click on the map.
	 * 
	 * @param panelCoordinates
	 *            Click coordinates, in pixels on this panel!
	 */
	private void notifyWorldClickListeners(MouseEvent mouseEvent) {
		Pair<Integer> panelCoordinates = new Pair<Integer>(mouseEvent.getX(),
				mouseEvent.getY());
		Pair<Integer> gridCoordinates = panelCoordinates2Grid(panelCoordinates);
		Pair<Double> worldCoordinates;

		// This is synchronized so both calls are guaranteed to refer to the
		// same slam map
		synchronized (this) {
			double worldX = slamMap.idx2x(gridCoordinates.getFirst(),
					LengthUnit.METER);
			double worldY = slamMap.idx2y(gridCoordinates.getSecond(),
					LengthUnit.METER);

			worldCoordinates = new Pair<Double>(worldX, worldY);
		}

		// Notify all listeners
		List<WorldClickListener> listenersCopy = new ArrayList<WorldClickListener>(
				listeners);
		for (WorldClickListener listener : listenersCopy) {
			listener.worldClicked(worldCoordinates, mouseEvent);
		}

	}

	/**
	 * Adds a new listeners for clicks into the world.
	 * 
	 * @param listener
	 *            The new listener, should not be null.
	 */
	public void addWorldClickListener(WorldClickListener listener) {
		this.listeners.add(listener);
	}

	/**
	 * Removes a listener for clicks into the world.
	 * 
	 * @param listener
	 *            The listener to be removed, should not be null.
	 * @return <code>true</code>, if this listener was really registered.
	 */
	public boolean removeWorldClickListener(WorldClickListener listener) {
		return this.listeners.remove(listener);
	}

	public void updateAnnotations(HashMap<String, MapArea> areas,
			MapRegion currentRegion) {

		this.areas = areas;
		this.currentRegion = currentRegion;
		repaint();

	}

	/**
	 * A listeners for clicks on the world map.
	 * 
	 * @author jhelbach
	 */
	public static interface WorldClickListener {

		/**
		 * Called whenever the user has clicked on a location in the global map.
		 * 
		 * @param worldCoordinates
		 *            Global world position the user has clicked on, never null.
		 * @param mouseEvent
		 *            For getting things like the click count or which button
		 *            was clicked, never null.
		 */
		void worldClicked(Pair<Double> worldCoordinates, MouseEvent mouseEvent);
	}

}
