/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.util.xml;

import de.unibi.airobots.bonsai.util.helper.annotation.MapArea;
import de.unibi.airobots.bonsai.util.helper.annotation.MapPoint;
import de.unibi.airobots.bonsai.util.helper.annotation.MapRegion;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData.GoalType;
import de.unibi.airobots.btl.units.AngleUnit;
import de.unibi.airobots.btl.units.LengthUnit;

import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;
import java.io.IOException;
import java.util.List;
import javax.vecmath.Vector3d;
import nu.xom.*;
import org.jaxen.JaxenException;
import org.jaxen.xom.XOMXPath;
import org.xml.sax.SAXException;

/**
 * A couple of convenience methods for common parsing tasks.
 * 
 * @author Ingo Luetkebohle <iluetkeb@techfak.uni-bielefeld.de>
 */
public class XMLDecoderUtil {

    /**
     * Retrieve attribute value, optionally checking whether it is present.
     * 
     * @param e
     *            Element that contains the attribute
     * @param attrName
     *            name of attribute
     * @param required
     *            if true, an error is thrown when the attribute is not present.
     *            otherwise, null is returned.
     * @throws ParameterMissingException
     *             when required is true and the attribute is not present
     */
    private static String getValueChecked(Element e, String attrName,
            boolean required) throws ParameterMissingException {
        String value = e.getAttributeValue(attrName);
        if (value == null && required) {
            throw new ParameterMissingException("Attribute " + attrName
                    + " not present in element " + e);
        }

        return value;
    }

    /**
     * Get value of named attribute as a double.
     * 
     * @param e
     *            Element to look up in
     * @param attrName
     *            name of the desired attribute
     * @param required
     *            whether a missing attribute is considered an error
     * @throws ParameterMissingException
     *             when required is true and the attribute is not present
     * @throws NumberFormatException
     *             when the value cannot be parsed as a double
     * @return parsed double
     */
    public static final double getDouble(Element e, String attrName,
            boolean required) throws ParameterMissingException,
            NumberFormatException {
        return Double.parseDouble(getValueChecked(e, attrName, required));
    }

    /**
     * Get value of named attribute as an integer.
     * 
     * @param e
     *            Element to look up in
     * @param attrName
     *            name of the desired attribute
     * @param required
     *            whether a missing attribute is considered an error
     * @throws ParameterMissingException
     *             when required is true and the attribute is not present
     * @throws NumberFormatException
     *             when the value cannot be parsed as an integer
     * @return parsed integer
     */
    public static final int getInt(Element e, String attrName, boolean required)
            throws ParameterMissingException, NumberFormatException {
        return Integer.parseInt(getValueChecked(e, attrName, required));
    }

    public static final MapArea getMapArea(Element e) throws NumberFormatException, ParameterMissingException {

        Attribute type = e.getAttribute(XMLEncoderUtil.TYPE_ATTRNAME);
        MapArea mapArea = new MapArea(MapArea.MapAreaType.valueOf(type.getValue()));

        Attribute name = e.getAttribute(XMLEncoderUtil.NAME_ATTRNAME);
        mapArea.setArenaName(name.getValue());

        Elements child = e.getChildElements();

        for (int i = 0; i < child.size(); i++) {
            mapArea.addRegion(getMapRegion(child.get(i)));
        }

        return mapArea;
    }

    private static MapRegion getMapRegion(Element e) throws NumberFormatException, ParameterMissingException {

        Attribute type = e.getAttribute(XMLEncoderUtil.TYPE_ATTRNAME);
        MapRegion mapRegion = new MapRegion(type.getValue());

        Attribute label = e.getAttribute(XMLEncoderUtil.LABEL_ATTRNAME);
        mapRegion.setLabel(label.getValue());

        Attribute group = e.getAttribute(XMLEncoderUtil.GROUP_ATTRNAME);
        mapRegion.setGroup(group.getValue());
        
        try {
            Attribute high = e.getAttribute(XMLEncoderUtil.HIGH_ATTRNAME);
            mapRegion.setHeight(Double.parseDouble(high.getValue()));
        } catch (Exception ex) {
            //hack to be compatible with old xml versions
            mapRegion.setHeight(0);
        }

        Elements child = e.getChildElements();

        for (int i = 0; i < child.size(); i++) {
            
            String name = child.get(i).getQualifiedName();
            
            if (name.equals(XMLEncoderUtil.COORD_ELEMENTNAME)) {
                mapRegion.addPoint(getMapPoint(child.get(i)));
            } else {
                mapRegion.setViewPoint(getViewPoint(child.get(i)));
            }
        }

        return mapRegion;
    }

    private static NavigationGoalData getViewPoint(Element e) throws NumberFormatException, ParameterMissingException {
        
        double x = getDouble(e, XMLEncoderUtil.X_ATTRNAME, true);
        double y = getDouble(e, XMLEncoderUtil.Y_ATTRNAME, true);
        double yaw = getDouble(e, XMLEncoderUtil.YAW_ATTRNAME, true);
        NavigationGoalData vp = new NavigationGoalData("XMLDecoderUtil", x ,y, yaw, GoalType.GLOBAL,LengthUnit.METER,AngleUnit.RADIAN);
        
        return vp;
    }

    private static MapPoint getMapPoint(Element e) throws NumberFormatException, ParameterMissingException {

        double x = getDouble(e, XMLEncoderUtil.X_ATTRNAME, true);
        double y = getDouble(e, XMLEncoderUtil.Y_ATTRNAME, true);
        Attribute kind = e.getAttribute(XMLEncoderUtil.KIND_ATTRNAME);
        Attribute unit = e.getAttribute(XMLEncoderUtil.UNIT_ATTRNAME);
        Attribute ref = e.getAttribute(XMLEncoderUtil.REF_ATTRNAME);

        MapPoint point =
                new MapPoint(x, y, MapPoint.Kind.valueOf(kind.getValue()),
                        MapPoint.Unit.valueOf(unit.getValue()), ref.getValue());

        return point;
    }

    public static final Polygon getPolygone(Element e, boolean required)
            throws NumberFormatException, ParameterMissingException {

        Polygon polygon = new Polygon();

        Elements child = e.getChildElements();

        for (int i = 0; i < child.size(); i++) {
            polygon.addPoint(getInt(child.get(i), XMLEncoderUtil.X_ATTRNAME,
                    required), getInt(child.get(i), XMLEncoderUtil.Y_ATTRNAME,
                    required));

        }

        return polygon;
    }

    /**
     * Parses integer rectangle from the given element by looking at the
     * attributes "x", "y", "width" and "height".
     * 
     * @param e
     *            element to parse rectangle from
     * @param required
     *            whether a missing attribute is considered an error
     * @return parsed rectangle
     * @throws ParameterMissingException
     *             parsing error
     */
    public static final Rectangle getRectangle(Element e, boolean required)
            throws ParameterMissingException {
        return getRectangleWH(e, required, XMLEncoderUtil.X_ATTRNAME,
                XMLEncoderUtil.Y_ATTRNAME, XMLEncoderUtil.W_ATTRNAME,
                XMLEncoderUtil.H_ATTRNAME);
    }

    /**
     * Parses integer rectangle specified as upper-left, widht-height from the
     * given element by looking at the specified attributes.
     * 
     * @param e
     *            Element to look in
     * @param required
     *            whether a missing attribute is considered an error
     * @param xName
     *            Name of the attribute that contains the upper-left x
     * @param yName
     *            Name of the attribute that contains the upper-left y
     * @param wName
     *            Name of the attribute that contains the width
     * @param hName
     *            Name of the attribute that contains height
     * @return parsed rectangle
     * @throws ParameterMissingException
     *             parsing error
     */
    public static final Rectangle getRectangleWH(Element e, boolean required,
            String xName, String yName, String wName, String hName)
            throws ParameterMissingException {
        return new Rectangle(getInt(e, xName, required), getInt(e, yName,
                required), getInt(e, wName, required), getInt(e, hName,
                required));
    }

    /**
     * Get a rectangle specified by the upper-left and lower-right corners.
     * 
     * @param e
     *            element to parse from
     * @param required
     *            whether a missing attribute is considered an error
     * @param x1Name
     *            name of the attribute containing the x1 coordinate
     * @param x2Name
     *            name of the attribute containing the x2 coordinate
     * @param y1Name
     *            name of the attribute containing the y1 coordinate
     * @param y2Name
     *            name of the attribute containing the y2 coordinate
     * @return parsed rectangle
     * @throws ParameterMissingException
     *             parsing error
     */
    public static final Rectangle getRectangleULLR(Element e, boolean required,
            String x1Name, String y1Name, String x2Name, String y2Name)
            throws ParameterMissingException {
        final int x1 = getInt(e, x1Name, required);
        final int y1 = getInt(e, y2Name, required);
        return new Rectangle(x1, y1, getInt(e, x2Name, required) - x1, getInt(
                e, y2Name, required)
                - y1);
    }

    /**
     * Parses double rectangle from the given element by looking at the
     * attributes "x", "y", "width" and "height".
     * 
     * @param e
     *            element to parse from
     * @param required
     *            whether a missing attribute is considered an error
     * @return parsed rectangle
     * @throws ParameterMissingException
     *             parsing error
     */
    public static final Rectangle2D.Double getRectangle2D(Element e,
            boolean required) throws ParameterMissingException {
        return new Rectangle2D.Double(getDouble(e, "x", required), getDouble(e,
                "y", required), getDouble(e, "width", required), getDouble(e,
                "height", required));
    }

    /**
     * Parses double world pos from the given element by looking at the
     * attributes "x", "y" and "z".
     * 
     * @param e
     *            element to parse from
     * @param required
     *            whether a missing attribute is considered an error
     * @return parsed vector
     * @throws ParameterMissingException
     *             parsing error
     */
    public static final Vector3d getWorldPos(Element e, boolean required)
            throws ParameterMissingException {
        return new Vector3d(getDouble(e, "x", required), getDouble(e, "y",
                required), getDouble(e, "z", required));
    }

    /**
     * Parses a string from an element and an attribute name.
     * 
     * @param e
     *            element to parse string from
     * @param attrName
     *            name of the attribute containing the string
     * @param required
     *            whether a missing attribute is considered an error
     * @return parsed string
     * @throws ParameterMissingException
     *             parsing error
     */
    public static final String getString(Element e, String attrName,
            boolean required) throws ParameterMissingException {
        return getValueChecked(e, attrName, required);
    }

    public static final Element getFirstElementFromQuery(Element root,
            String xpath) {
        Nodes nodes = root.query(xpath);
        for (int i = 0; i < nodes.size(); i++) {
            if (nodes.get(i) instanceof Element) {
                return (Element) (nodes.get(i));
            }
        }
        return null;
    }

    public static final Attribute getFirstAttributeFromQuery(Element root,
            String xpath) {
        Nodes nodes = root.query(xpath);
        for (int i = 0; i < nodes.size(); i++) {
            if (nodes.get(i) instanceof Attribute) {
                return (Attribute) (nodes.get(i));
            }
        }
        return null;
    }

    public static final Node getFirstNodeFromQuery(Element root, String xpath)
            throws ParameterMissingException {
        Nodes nodes = root.query(xpath);
        if (nodes.size() < 1) {
            throw new ParameterMissingException("No nodes found at" + xpath);
        }
        return nodes.get(0);
    }

    public static final Node getFirstNodeFromQuery(Element root, XOMXPath xpath)
            throws ParameterMissingException, JaxenException {
        List<?> nodes = xpath.selectNodes(root);
        if (nodes.size() < 1) {
            throw new ParameterMissingException("No nodes found at" + xpath);
        }
        return (Node) nodes.get(0);
    }

    public static final Element loadXML(String path) throws SAXException,
            ValidityException, ParsingException, IOException {

        Builder builder = new Builder();
        Document doc = builder.build(path);
        return doc.getRootElement();
    }

}
