/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.xcf;

import net.sf.xcf.util.QueueClosedException;
import net.sf.xcf.util.SynchronizedQueue;

/**
 * Creates a synchronized bound queue with a limited capacity. Before pushing an
 * object into the queue it removes elements so that the size of the queue is
 * always within the capacity.
 * 
 * @author marc
 * @author jwienke
 * @param <T>
 *            type of elements contained in this queue
 */
public abstract class BoundSynchronizedQueue<T> extends SynchronizedQueue<T> {

    private int capacity;

    /**
     * Returns the maximum capacity of this queue.
     * 
     * @return the capacity
     */
    public int getCapacity() {
        return capacity;
    }

    /**
     * Sets the maximum capacity of this queue.
     * 
     * @param capacity
     *            the capacity to set
     */
    public void setCapacity(int capacity) {
        this.capacity = capacity;
        // TODO what if this queue contains more entries than the new capacity?
    }

    /**
     * Create queue with a given capacity.
     * 
     * @param capacity
     *            the initial capacity.
     */
    public BoundSynchronizedQueue(int capacity) {
        super();
        this.capacity = capacity;
    }

    /**
     * Create queue with a capacity of 1. This implements a synchronized buffer.
     */
    public BoundSynchronizedQueue() {
        super();
        this.capacity = 1;
    }

    @Override
    public synchronized void push(T element) {
        if (super.isClosed()) {
            throw new QueueClosedException("Queue is closed!");
        }
        // discard all older elements

        while (super.getSize() >= capacity) {
            super.pop();
        }
        super.push(element);
    }

}
