/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.xcf.actuators;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;
import net.sf.xcf.transport.XOPData;
import nu.xom.Attribute;
import nu.xom.Document;
import nu.xom.Element;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.OldCameraActuator;
import de.unibi.airobots.bonsai.actuators.data.CameraPosition;
import de.unibi.airobots.bonsai.xcf.RemoteServerReconnectable;

/**
 * Actuator which can be used to access the Pan-Tilt-Unit of the Camera using
 * the old C++ camserver.
 * 
 * @author mhanheid
 * @author jwienke
 */
public class XcfOldCameraActuator implements OldCameraActuator,
        RemoteServerReconnectable {

    private Logger logger = Logger.getLogger(getClass());

    private String remoteServerName = "";

    private Object remoteServerLock = new Object();
    private RemoteServer remoteServer = null;

    /**
     * Constructor.
     * 
     * @param remoteServerName
     *            name of the remote server used to control the camera
     */
    public XcfOldCameraActuator(String remoteServerName) {
        this.remoteServerName = remoteServerName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setPos(double pan, double tilt) throws IOException {

        synchronized (remoteServerLock) {
            if (remoteServer == null) {
                throw new IllegalStateException("No remote server set.");
            }
        }

        try {

            Element camposition = new Element("CAMPOSITION");
            Document doc = new Document(camposition);
            if (pan > Math.PI) {
                pan = Math.PI - pan;
            }
            if (tilt > Math.PI) {
                tilt = Math.PI - tilt;
            }

            camposition
                    .addAttribute(new Attribute("pan", Double.toString(pan)));
            camposition.addAttribute(new Attribute("tilt", Double
                    .toString(tilt)));

            synchronized (remoteServerLock) {
                this.remoteServer.callMethod("vSetPos", new XOPData(doc));
            }

        } catch (MethodNotFoundException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (UserException e) {
            logger.error(e);
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setZoom(double zoom) throws IOException {

        synchronized (remoteServerLock) {
            if (remoteServer == null) {
                throw new IllegalStateException("No remote server set.");
            }
        }

        try {

            Element camposition = new Element("CAMZOOM");
            Document doc = new Document(camposition);
            camposition.addAttribute(new Attribute("zoom", Double
                    .toString(zoom)));

            XOPData res;
            synchronized (remoteServerLock) {
                res =
                        this.remoteServer.callMethod("vSetZoom", new XOPData(
                                doc));
            }
            res.getDocument();

        } catch (MethodNotFoundException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (UserException e) {
            logger.error(e);
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setPanSpeed(double panspeed) throws IOException {

        synchronized (remoteServerLock) {
            if (remoteServer == null) {
                throw new IllegalStateException("No remote server set.");
            }
        }

        try {

            Element camposition = new Element("CAMSPEED");
            Document doc = new Document(camposition);
            camposition.addAttribute(new Attribute("ps", Double
                    .toString(panspeed)));
            XOPData res;
            synchronized (remoteServerLock) {
                res =
                        this.remoteServer.callMethod("vSetPanSpeed",
                                new XOPData(doc));
            }
            res.getDocument();

        } catch (MethodNotFoundException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (UserException e) {
            logger.error(e);
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setTiltSpeed(double tiltspeed) throws IOException {

        synchronized (remoteServerLock) {
            if (remoteServer == null) {
                throw new IllegalStateException("No remote server set.");
            }
        }

        try {

            Element camposition = new Element("CAMSPEED");
            Document doc = new Document(camposition);
            camposition.addAttribute(new Attribute("ts", Double
                    .toString(tiltspeed)));

            XOPData res;
            synchronized (remoteServerLock) {
                res =
                        this.remoteServer.callMethod("vSetTiltSpeed",
                                new XOPData(doc));
            }
            res.getDocument();

        } catch (MethodNotFoundException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (UserException e) {
            logger.error(e);
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setZoomSpeed(double zoomspeed) throws IOException {

        synchronized (remoteServerLock) {
            if (remoteServer == null) {
                throw new IllegalStateException("No remote server set.");
            }
        }

        try {

            Element camposition = new Element("CAMSPEED");
            Document doc = new Document(camposition);
            camposition.addAttribute(new Attribute("zs", Double
                    .toString(zoomspeed)));

            XOPData res;
            synchronized (remoteServerLock) {
                res =
                        this.remoteServer.callMethod("vSetZoomSpeed",
                                new XOPData(doc));
            }
            res.getDocument();

        } catch (MethodNotFoundException e) {
            logger.error(e);
            throw new IOException(e);
        } catch (UserException e) {
            logger.error(e);
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public float getZoomPos() throws IOException {

        synchronized (remoteServerLock) {
            if (remoteServer == null) {
                throw new IllegalStateException("No remote server set.");
            }
        }

        try {

            Element zoom = new Element("ZOOMPOS");
            Document doc = new Document(zoom);
            zoom.addAttribute(new Attribute("zoomPos", "1"));

            XOPData res;
            synchronized (remoteServerLock) {
                res =
                        this.remoteServer.callMethod("fGetZoomPos",
                                new XOPData(doc));
            }
            return Float.parseFloat(res.getDocument().getRootElement()
                    .getAttributeValue("zoomPos"));

        } catch (Exception e) {
            logger.error(e);
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public CameraPosition getCameraPosition() throws IOException,
            MethodNotFoundException, UserException {

        synchronized (remoteServerLock) {
            if (remoteServer == null) {
                throw new IllegalStateException("No remote server set.");
            }
        }

        Element pos = new Element("CAMPOSITION");
        Document doc = new Document(pos);

        pos.addAttribute(new Attribute("tilt", "1"));
        pos.addAttribute(new Attribute("pan", "1"));

        double tilt;
        double pan;
        double zoom;

        XOPData res = this.remoteServer.callMethod("vGetPos", new XOPData(doc));

        pan =
                Double.parseDouble(res.getDocument().getRootElement()
                        .getAttributeValue("pan"));
        tilt =
                Double.parseDouble(doc.getRootElement().getAttributeValue(
                        "tilt"));
        zoom = this.getZoomPos();

        synchronized (remoteServerLock) {
            return new CameraPosition(pan, tilt, zoom, System
                    .currentTimeMillis());
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);
        return servers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetRemoteServer(RemoteServer remoteServer) {
        assert remoteServer.getServerName().equals(remoteServerName);
        synchronized (remoteServerLock) {
            this.remoteServer = remoteServer;
        }
    }

}
